# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, Dict, cast

from discopop_explorer.PEGraphX import (
    CUNode,
    FunctionNode,
    LoopNode,
    MWType,
    EdgeType,
    Node,
    PEGraphX,
    LineID,
)
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import (
    recursive_function_call_contained_in_worker_cu,
    line_contained_in_region,
    contains_reduction,
)
from discopop_explorer.utils import classify_task_vars


def detect_task_suggestions(pet: PEGraphX) -> List[PatternInfo]:
    """creates task parallelism suggestions and returns them as a list of
    TaskParallelismInfo objects.
    Currently relies on previous processing steps and suggests WORKER CUs
    as Tasks and BARRIER/BARRIER_WORKER as Taskwaits.
    :param pet: PET graph
    :return: List[TaskParallelismInfo]
    """
    # suggestions contains a map from LID to a set of suggestions. This is required to
    # detect multiple suggestions for a single line of source code.
    suggestions: Dict[str, List[TaskParallelismInfo]] = dict()  # LID -> List[TaskParallelismInfo]

    # get a list of cus classified as WORKER
    worker_cus: List[Node] = []
    barrier_cus: List[Node] = []
    barrier_worker_cus: List[Node] = []

    func_cus: List[Node] = []

    for v in pet.all_nodes():
        if v.mw_type == MWType.WORKER:
            worker_cus.append(v)
        if v.mw_type == MWType.BARRIER:
            barrier_cus.append(v)
        if v.mw_type == MWType.BARRIER_WORKER:
            barrier_worker_cus.append(v)
        if isinstance(v, FunctionNode):
            func_cus.append(v)

    worker_cus = worker_cus + barrier_worker_cus + func_cus

    # SUGGEST TASKWAIT
    for v in barrier_cus:
        # get line number of first dependency. suggest taskwait prior to that
        first_dependency_line = v.end_position()
        first_dependency_line_number = first_dependency_line[first_dependency_line.index(":") + 1 :]
        for s, t, e in pet.out_edges(v.id):
            if e.etype == EdgeType.DATA:
                dep_line = cast(LineID, e.sink_line)
                dep_line_number = dep_line[dep_line.index(":") + 1 :]
                if dep_line_number < first_dependency_line_number:
                    first_dependency_line = dep_line
        tmp_suggestion = TaskParallelismInfo(v, TPIType.TASKWAIT, ["taskwait"], first_dependency_line, [], [], [])
        if v.start_position() not in suggestions:
            # no entry for source code line contained in suggestions
            suggestions[v.start_position()] = []
            suggestions[v.start_position()].append(tmp_suggestion)
        else:
            # entry for source code line already contained in suggestions
            suggestions[v.start_position()].append(tmp_suggestion)

    # SUGGEST TASKS
    for vx in pet.all_nodes():
        # iterate over all entries in recursiveFunctionCalls
        # in order to find task suggestions
        for i in range(0, len(vx.recursive_function_calls)):
            function_call_string = vx.recursive_function_calls[i]
            if not type(function_call_string) == str:
                continue
            contained_in = recursive_function_call_contained_in_worker_cu(function_call_string, worker_cus)
            if contained_in is not None:
                current_suggestions = None
                # recursive Function call contained in worker cu
                # -> issue task suggestion
                pragma_line = function_call_string[function_call_string.index(":") + 1 :]
                pragma_line = pragma_line.replace(",", "").replace(" ", "")

                # only include cu and func nodes
                if not isinstance(contained_in, (FunctionNode, CUNode)):
                    continue
                if (
                    contained_in.mw_type == MWType.WORKER
                    or contained_in.mw_type == MWType.BARRIER_WORKER
                    or isinstance(contained_in, FunctionNode)
                ):
                    # suggest task
                    fpriv, priv, shared, in_dep, out_dep, in_out_dep, red = classify_task_vars(
                        pet, contained_in, "", [], [], used_in_task_parallelism_detection=True
                    )
                    current_suggestions = TaskParallelismInfo(
                        vx,
                        TPIType.TASK,
                        ["task"],
                        pragma_line,
                        [v.name for v in fpriv],
                        [v.name for v in priv],
                        [v.name for v in shared],
                    )

                # insert current_suggestions into suggestions
                # check, if current_suggestions contains an element
                if current_suggestions is not None:
                    # current_suggestions contains something
                    if pragma_line not in suggestions:
                        # LID not contained in suggestions
                        suggestions[pragma_line] = []
                        suggestions[pragma_line].append(current_suggestions)
                    else:
                        # LID already contained in suggestions
                        suggestions[pragma_line].append(current_suggestions)
    # end of for loop

    # construct return value (list of TaskParallelismInfo)
    result: List[PatternInfo] = []
    for key in suggestions:
        for single_suggestion in suggestions[key]:
            result.append(single_suggestion)
    return result


def correct_task_suggestions_in_loop_body(pet: PEGraphX, suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """Separate treatment of task suggestions at loop increment CUs.
    If regular loop: move taskwait suggested at loop increment line to end of loop body.
    If do-all loop: move taskwait suggested at loop increment line outside of loop body.
    If critical CUs are detected in the process, they will be appended to the critical_sections
    of the respective task suggestion.
    If atomic CUs are detected in the process, they will be appended to the atomic_sections
    of the respective task suggestion.
    :param pet: PET graph
    :param suggestions: Found suggestions
    :return: Updated suggestions"""
    task_suggestions = [s for s in [e for e in suggestions if type(e) == TaskParallelismInfo] if s.type is TPIType.TASK]
    for ts in task_suggestions:
        found_critical_cus: List[Node] = []
        found_atomic_cus: List[Node] = []
        for loop_cu in pet.all_nodes(LoopNode):
            # check if task suggestion inside do-all loop exists
            if line_contained_in_region(ts._node.start_position(), loop_cu.start_position(), loop_cu.end_position()):

                def find_taskwaits(cu_node: Node, visited: List[Node]):
                    if cu_node.tp_contains_taskwait:
                        return [cu_node]
                    result = []
                    visited.append(cu_node)
                    for succ_cu_node in [
                        pet.node_at(t)
                        for s, t, e in pet.out_edges(cu_node.id)
                        if e.etype == EdgeType.SUCCESSOR and pet.node_at(t) != cu_node
                    ]:
                        if succ_cu_node not in visited:
                            result += find_taskwaits(succ_cu_node, visited)
                    return result

                # find successive taskwaits
                successive_taskwait_cus = find_taskwaits(ts._node, [])
                for stws_cu in successive_taskwait_cus:
                    if loop_cu.do_all:
                        # check if stws is suggested at loop increment
                        if stws_cu.basic_block_id != "for.inc":
                            continue
                        # Do-all loop, move taskwait to the outside
                        print(
                            "TPDet: correct_task_suggestions_in_loop_body: Task in do-all loop ",
                            ts.pragma_line,
                            ". Moving Taskwait ",
                            stws_cu.start_line,
                            " to: ",
                            int(loop_cu.end_position().split(":")[1]) + 1,
                        )
                        for s in suggestions:
                            if type(s) == TaskParallelismInfo:
                                if s.type is TPIType.TASKWAIT and s._node == stws_cu:
                                    s.pragma_line = int(loop_cu.end_position().split(":")[1]) + 1
                    else:
                        # Regular loop: task = loop body, move taskwait to the end of the loop body
                        # protect RAW to shared with critical section around CU (general)  or atomic (reduction)
                        print(
                            "TPDet: correct_task_suggestions_in_loop_body: Task in regular loop ",
                            ts.pragma_line,
                            ". Moving Taskwait ",
                            stws_cu.start_line,
                            " to: ",
                            int(loop_cu.end_position().split(":")[1]),
                            ".",
                        )
                        # move pragma taskwait line
                        for s in suggestions:
                            if type(s) == TaskParallelismInfo:
                                if s.type is TPIType.TASKWAIT and s._node == stws_cu:
                                    s.pragma_line = int(loop_cu.end_position().split(":")[1])
                        # move pragma task line to beginning of loop body (i.e. make the entire loop body a task)
                        # set task region lines accordingly
                        # if ts._node is a direct child of loop_cu
                        if loop_cu.id in [e[0] for e in pet.in_edges(ts._node.id, EdgeType.CHILD)]:
                            print(
                                "Moving Pragma from: ",
                                ts.pragma_line,
                                " to: ",
                                int(loop_cu.start_position().split(":")[1]) + 1,
                            )
                            ts.pragma_line = int(loop_cu.start_position().split(":")[1]) + 1
                            ts.region_start_line = str(ts.pragma_line)
                            ts.region_end_line = loop_cu.end_position().split(":")[1]

                            # protect RAW-Writes to shared variables with critical section
                            # i.e. find in-deps to shared variables and suggest critical section around CUs
                            # containing such cases
                            for loop_cu_child in pet.direct_children_or_called_nodes(loop_cu):
                                for in_dep_var_name in list(
                                    set([e[2].var_name for e in pet.in_edges(loop_cu_child.id, EdgeType.DATA)])
                                ):
                                    if in_dep_var_name in ts.shared:
                                        # check if the found dependency occurs in the scope of the suggested task
                                        if (
                                            loop_cu_child.file_id == ts._node.file_id
                                            and loop_cu_child.start_line >= int(ts.region_start_line)
                                            and loop_cu_child.end_line <= int(ts.region_end_line)
                                        ):
                                            # seperate between critical and atomic CUs
                                            if contains_reduction(pet, loop_cu_child):
                                                # split cu lines on reduction, mark surrounding lines as critical
                                                file_idx = loop_cu_child.start_position().split(":")[0]
                                                start_line = int(loop_cu_child.start_position().split(":")[1])
                                                end_line = int(loop_cu_child.end_position().split(":")[1])
                                                critical_lines_range = range(start_line, end_line + 1)
                                                atomic_lines = []
                                                # seperate between critical and atomic lines
                                                for red_var in pet.reduction_vars:
                                                    if line_contained_in_region(
                                                        red_var["reduction_line"],
                                                        loop_cu_child.start_position(),
                                                        loop_cu_child.end_position(),
                                                    ):
                                                        atomic_lines.append(
                                                            int(red_var["reduction_line"].split(":")[1])
                                                        )
                                                critical_lines = [
                                                    e for e in critical_lines_range if e not in atomic_lines
                                                ]
                                                # combine successive critical lines if possible
                                                combined_critical_lines = [
                                                    (e, e) for e in critical_lines
                                                ]  # (start, end)
                                                found_combination = True
                                                while found_combination and len(combined_critical_lines) > 1:
                                                    found_combination = False
                                                    for outer_idx, outer in enumerate(combined_critical_lines):
                                                        for inner_idx, inner in enumerate(combined_critical_lines):
                                                            if inner_idx == outer_idx:
                                                                continue
                                                            if outer[1] + 1 == inner[0]:
                                                                # inner is direct successor of outer
                                                                combined_critical_lines[outer_idx] = (
                                                                    outer[0],
                                                                    inner[1],
                                                                )
                                                                combined_critical_lines.pop(inner_idx)
                                                                found_combination = True
                                                                break
                                                        if found_combination:
                                                            break
                                                # append critical and atomic to ts.atomic_sections/ts.critical_sections
                                                for e in combined_critical_lines:
                                                    ts.critical_sections.append(
                                                        ""
                                                        + str(file_idx)
                                                        + ":"
                                                        + str(e[0])
                                                        + "-"
                                                        + str(file_idx)
                                                        + ":"
                                                        + str(e[1])
                                                    )
                                                for a in atomic_lines:
                                                    ts.atomic_sections.append(
                                                        ""
                                                        + str(file_idx)
                                                        + ":"
                                                        + str(a)
                                                        + "-"
                                                        + str(file_idx)
                                                        + ":"
                                                        + str(a)
                                                    )
                                            else:
                                                # append loop_cu_child to list of critical CUs
                                                found_critical_cus.append(loop_cu_child)
        # CRITICAL SECTIONS
        __identify_atomic_or_critical_sections(pet, ts, found_critical_cus, False)
        # ATOMIC SECTIONS
        __identify_atomic_or_critical_sections(pet, ts, found_atomic_cus, True)
    return suggestions


def __identify_atomic_or_critical_sections(
    pet: PEGraphX, ts: TaskParallelismInfo, found_cus: List[Node], selector: bool
):
    """Identifies and marks atomic or critical sections.
    :param pet: PET Graph
    :param ts: task suggestion
    :param found_cus: list of previously identified atomic or critical cus.
    :param selector: True: identify atomic sections. False: identify critical sections
    """
    # remove potential duplicates from found cus
    found_cus = list(set(found_cus))
    # get lists of combinable cus by checking successor relation
    combinations = []
    for cu in found_cus:
        combinations.append([cu])
    found_combination = True
    while found_combination:
        found_combination = False
        for parent_idx in range(0, len(combinations)):
            if found_combination:
                break
            for child_idx in range(0, len(combinations)):
                if found_combination:
                    break
                if parent_idx == child_idx:
                    continue
                if combinations[child_idx][0] in pet.direct_successors(combinations[parent_idx][-1]):
                    combinations[parent_idx] += combinations[child_idx]
                    combinations.pop(child_idx)
                    found_combination = True
    # remove entries from combinations, if they are already covered by another combination
    # occurs, if line numbers are overlapping although CUs are not direct successors of each other.
    removed_entry = True
    while removed_entry:
        removed_entry = False
        for parent_idx in range(0, len(combinations)):
            if removed_entry:
                break
            # check that parent is a single-entry list
            if len(combinations[parent_idx]) != 1:
                continue
            for child_idx in range(0, len(combinations)):
                if removed_entry:
                    break
                if parent_idx == child_idx:
                    continue
                # check if parent is covered by child
                parent = combinations[parent_idx]
                child = combinations[child_idx]
                if line_contained_in_region(
                    parent[0].start_position(), child[0].start_position(), child[-1].end_position()
                ) and line_contained_in_region(
                    parent[-1].end_position(), child[0].start_position(), child[-1].end_position()
                ):
                    combinations.pop(parent_idx)
                    removed_entry = True
    # create a string from the gathered information and append to ts.critical_sections
    for combination_list in combinations:
        section_str = ""
        section_str += combination_list[0].start_position()
        section_str += "-"
        section_str += combination_list[-1].end_position()
        if selector:
            ts.atomic_sections.append(section_str)
        else:
            ts.critical_sections.append(section_str)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Optional


class Variable(object):
    operation: Optional[str]

    def __init__(self, type, name, defLine, accessMode="", sizeInByte=0):
        self.type = type
        self.name: str = name
        self.defLine = defLine
        self.operation = None
        self.accessMode = accessMode
        # To prevent potentially dividing by 0, set the sizeInByte to 8 in case it is unknown.
        # Sizes are mainly unknown if the underlying variable is a pointer type.
        # Since Pointers in 64-bit architectures occupy 8 bytes, this size is chosen as the "default" at this point.
        if sizeInByte == 0:
            sizeInByte = 8
        self.sizeInByte = sizeInByte

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return isinstance(other, Variable) and self.name == other.name

    def __lt__(self, other):
        return self.name < other.name

    def __str__(self):
        return self.name

    def toJSON(self):
        if self.operation is None:
            return self.name
        else:
            return self.operation + ":" + self.name



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Tuple, Set, cast

from discopop_explorer.PEGraphX import NodeID, PEGraphX, MemoryRegion, CUNode, EdgeType, DepType
from discopop_library.discopop_optimizer.classes.types.DataAccessType import (
    WriteDataAccess,
    ReadDataAccess,
)

next_free_unique_write_id: int = 0


def get_next_free_unique_write_id() -> int:
    global next_free_unique_write_id
    buffer = next_free_unique_write_id
    next_free_unique_write_id += 1
    return buffer


def get_data_accesses_for_cu(pet: PEGraphX, cu_id: NodeID) -> Tuple[Set[WriteDataAccess], Set[ReadDataAccess]]:
    """Calculates and returns the sets of accessed memory regions for the given cu node.
    The first element contains write accesses, the second element contains read accesses."""
    parent_function = pet.get_parent_function(pet.node_at(cu_id))
    subtree = pet.subtree_of_type(parent_function, CUNode)

    in_dep_edges = pet.in_edges(cu_id, EdgeType.DATA)
    out_dep_edges = pet.out_edges(cu_id, EdgeType.DATA)

    written_memory_regions = [
        WriteDataAccess(MemoryRegion(cast(str, d.memory_region)), get_next_free_unique_write_id(), d.var_name)
        for s, t, d in in_dep_edges
        if (d.dtype == DepType.RAW or d.dtype == DepType.WAW)
        and d.memory_region is not None
        and len(d.memory_region) != 0
    ]
    written_memory_regions += [
        WriteDataAccess(MemoryRegion(cast(str, d.memory_region)), get_next_free_unique_write_id(), d.var_name)
        for s, t, d in out_dep_edges
        if (d.dtype == DepType.WAR or d.dtype == DepType.WAW)
        and d.memory_region is not None
        and len(d.memory_region) != 0
    ]

    read_memory_regions = [
        ReadDataAccess(MemoryRegion(cast(str, d.memory_region)), d.var_name)
        for s, t, d in in_dep_edges
        if (d.dtype == DepType.WAR) and d.memory_region is not None and len(d.memory_region) != 0
    ]
    read_memory_regions += [
        ReadDataAccess(MemoryRegion(cast(str, d.memory_region)), d.var_name)
        for s, t, d in out_dep_edges
        if (d.dtype == DepType.RAW) and d.memory_region is not None and len(d.memory_region) != 0
    ]
    # remove duplicates (manual, since hashing MemoryRegions did not work as intended. might be fixed in the future)
    rmr_set: Set[ReadDataAccess] = set()
    for rmr in read_memory_regions:
        if str(rmr) not in [str(e) for e in rmr_set]:
            rmr_set.add(rmr)

    return set(written_memory_regions), rmr_set



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, Tuple, Union

import sympy
from extrap.entities.callpath import Callpath  # type: ignore
from extrap.entities.metric import Metric  # type: ignore
from extrap.entities.model import Model  # type: ignore
from extrap.entities.parameter import Parameter  # type: ignore
from extrap.fileio.json_file_reader import read_json_file  # type: ignore
from extrap.modelers.model_generator import ModelGenerator  # type: ignore
from extrap.modelers.multi_parameter.multi_parameter_modeler import MultiParameterModeler  # type: ignore
from sympy.parsing.sympy_parser import parse_expr  # type: ignore

from .Microbench import (
    Microbench,
    MicrobenchType,
    MicrobenchDimension,
    MicrobenchCoordinate,
)


# This class uses extrap to extrapolate microbench measurements.
class ExtrapInterpolatedMicrobench(Microbench):
    models: Dict[Tuple[Callpath, Metric], Model]

    def __init__(self, jsonFile: str):
        experiment = read_json_file(jsonFile)
        modeler = MultiParameterModeler()
        modeler.single_parameter_modeler.use_crossvalidation = False
        model_generator = ModelGenerator(experiment, modeler=modeler)  # TODO use_median??
        model_generator.model_all()
        self.models = experiment.modelers[0].models

    # convenient for debugging: print the model functions
    def printModelFunctions(self):
        for callpathMetric, model in self.models.items():
            test = str(callpathMetric[0]).ljust(12)
            metric = str(callpathMetric[1])
            if metric == "Test time in us":
                metric = "Test:      "
            elif metric == "Reference time in us":
                metric = "Reference: "
            elif metric == "Overhead time in us":
                metric = "Overhead:  "
            function = model.hypothesis.function.to_string()
            print(test, metric, function)

    def getFunctionString(
        self,
        benchType: MicrobenchType = MicrobenchType.DOALL,
        benchDim: MicrobenchDimension = MicrobenchDimension.OVERHEAD,
    ):
        return self.models[(Callpath(benchType), Metric(benchDim))].hypothesis.function.to_string()

    def getFunctionSympy(
        self,
        benchType: MicrobenchType = MicrobenchType.DOALL,
        benchDim: MicrobenchDimension = MicrobenchDimension.OVERHEAD,
    ):
        function_str = self.getFunctionString(benchType, benchDim)
        # NOTE: replacement order matters! "ab".replace("a","b").replace("b","a") --> "aa", NOT "ba"
        function_str = function_str.replace("^", "**")
        function_str = function_str.replace("r", "iterations")
        function_str = function_str.replace("p", "threads")
        function_str = function_str.replace("q", "workload")
        # define replacements to match representations used in extrap output
        function_mappings = {"log2": lambda x: sympy.log(x, 2)}
        expr = parse_expr(function_str, local_dict=function_mappings)
        return expr

    def getMeasurements(self):
        raise NotImplementedError("TODO")  # TODO

    def toJSON(self):
        raise NotImplementedError("TODO")  # TODO

    def evaluateInterpolation(
        self,
        benchType: MicrobenchType,
        benchDim: MicrobenchDimension,
        benchCoord: Union[MicrobenchCoordinate, Tuple[int, float, float]],
    ):
        return self.models[(Callpath(benchType), Metric(benchDim))].hypothesis.function.evaluate([*benchCoord])



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from sympy import Symbol  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.edges.GenericEdge import GenericEdge


class RequirementEdge(GenericEdge):
    """Used to represent dependencies between path selections in the form of:
    if A is selected, it is required that B is selected as well."""

    pass

    def get_cost_model(self) -> CostModel:
        raise ValueError("The cost of a RequirementEdge is not defined and may never be used!")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from .PEGraphX import Node, NodeID
from typing import List, Set

global_pet = None


def pet_function_metadata_initialize_worker(pet):
    global global_pet
    global_pet = pet


def pet_function_metadata_parse_func(func_node):
    stack: List[Node] = global_pet.direct_children(func_node)
    func_node.children_cu_ids = [node.id for node in stack]
    local_children: Set[NodeID] = set()

    while stack:
        child = stack.pop()
        local_children.add(child.id)
        children = global_pet.direct_children(child)
        func_node.children_cu_ids.extend([node.id for node in children])
        stack.extend(children)

    local_reachability_dict = func_node.calculate_reachability_pairs(global_pet)
    local_result = func_node.id, local_reachability_dict, local_children
    return local_result



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from enum import IntEnum


class ExitPointPositioning(IntEnum):
    BEFORE_CU = 0
    AFTER_CU = 1


class EntryPointPositioning(IntEnum):
    BEFORE_CU = 0
    AFTER_CU = 1


class ExitPointType(IntEnum):
    FROM_DEVICE = 0
    DELETE = 1
    ASYNC_FROM_DEVICE = 2


class EntryPointType(IntEnum):
    TO_DEVICE = 0
    ALLOCATE = 1
    ASYNC_TO_DEVICE = 2
    ASYNC_ALLOCATE = 3


class UpdateType(IntEnum):
    TO_DEVICE = 0
    FROM_DEVICE = 1
    # TO_FROM should not occur ideally, since data should only be modified either on the host or the device for now
    TO_FROM_DEVICE = 2
    ALLOCATE = 3



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import re
from typing import List, Iterator


def get_library(name: str) -> str:
    library_dirs: List[str]
    if os.getenv("DISCOPOP_INSTALL"):
        library_dirs = [
            os.path.expandvars("${DISCOPOP_INSTALL}"),
            os.path.expandvars("${DISCOPOP_INSTALL}/libi"),
            os.path.expandvars("${DISCOPOP_INSTALL}/rtlib"),
            os.path.expandvars("${DISCOPOP_INSTALL}/lib"),
        ]
    else:
        library_dirs = [
            os.path.expanduser("~/.local/lib"),
            "/usr/local/lib",
            "/usr/lib",
            "/lib",
        ]
    for library_dir in library_dirs:
        if os.path.exists(os.path.join(library_dir, name)):
            return os.path.join(library_dir, name)
    raise SystemExit(
        f"File {name} not found. Searched in: {', '.join(library_dirs)}.\n"
        f"Build DiscoPoP and set DISCOPOP_INSTALL environment variable."
    )


def is_compile(clang_args: List[str]) -> bool:
    return "-c" in clang_args or any([re.match(r"^[^-].+\.(?:c|cc|cpp)$", arg) for arg in clang_args])


def is_link(clang_args: List[str]) -> bool:
    return "-c" not in clang_args


def recursive_scandir(path: str) -> Iterator[os.DirEntry]:  # type: ignore
    with os.scandir(path) as dir_iter:
        for entry in dir_iter:
            yield entry
            if entry.is_dir():
                yield from recursive_scandir(entry.path)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import json
import os.path
from pathlib import Path
from typing import Dict

from discopop_library.LineMapping.load import load_line_mapping
from discopop_library.LineMapping.save import save_line_mapping


def initialize_line_mapping(
    file_mapping: Dict[int, Path],
    discopop_path: str = ".discopop",
):
    """initializes the line mapping dictionary to track line shifts due to inserted pragmas.
    The Dictionary will be stored in .discopop/line_mapping.json.
    Line ids start with 1."""
    line_mapping_dict: Dict[str, Dict[str, int]] = dict()

    # initialize line mapping (1->1, 2->2, ...)
    for file_id in file_mapping:
        if str(file_id) not in line_mapping_dict:
            line_mapping_dict[str(file_id)] = dict()
        line_id = 1
        with open(file_mapping[file_id], "r") as f:
            for line in f.readlines():
                line_mapping_dict[str(file_id)][str(line_id)] = line_id
                line_id += 1

    save_line_mapping(line_mapping_dict, discopop_path)

    # debug
    load_line_mapping(discopop_path)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter as tk
from tkinter import filedialog
import warnings

from discopop_library.PathManagement.PathManagement import get_path


def create_optimizer_screen(wizard, parent_frame, execution_configuration):
    # close elements on optimizer_frame
    for c in parent_frame.winfo_children():
        c.destroy()

    canvas = tk.Canvas(parent_frame)
    canvas.pack(fill=tk.BOTH)

    arguments = dict()

    def overwrite_with_file_selection(target: tk.Entry):
        prompt_result = filedialog.askopenfilename()
        if len(prompt_result) != 0:
            target.delete(0, tk.END)
            target.insert(0, prompt_result)

    ###
    tk.Label(
        canvas,
        text="Compile_command",
        justify=tk.RIGHT,
        anchor="e",
        font=wizard.style_font_bold_small,
    ).grid(row=1, column=0, sticky="ew")
    compile_command = tk.Entry(canvas, width=100)
    compile_command.insert(tk.END, "make")
    compile_command.grid(row=1, column=1, sticky="ew")
    ###
    tk.Label(
        canvas,
        text="DoAll microbench file",
        justify=tk.RIGHT,
        anchor="e",
        font=wizard.style_font_bold_small,
    ).grid(row=2, column=0, sticky="ew")
    doall_microbench_file = tk.Entry(canvas, width=100)
    doall_microbench_file.insert(tk.END, "None")
    doall_microbench_file.grid(row=2, column=1, sticky="ew")

    doall_microbench_file_path_selector = tk.Button(
        canvas, text="Select", command=lambda: overwrite_with_file_selection(doall_microbench_file)
    )
    doall_microbench_file_path_selector.grid(row=2, column=3)
    ###
    tk.Label(
        canvas,
        text="Reduction microbench file",
        justify=tk.RIGHT,
        anchor="e",
        font=wizard.style_font_bold_small,
    ).grid(row=3, column=0, sticky="ew")
    reduction_microbench_file = tk.Entry(canvas, width=100)
    reduction_microbench_file.insert(tk.END, "None")
    reduction_microbench_file.grid(row=3, column=1, sticky="ew")

    reduction_microbench_file_path_selector = tk.Button(
        canvas,
        text="Select",
        command=lambda: overwrite_with_file_selection(reduction_microbench_file),
    )
    reduction_microbench_file_path_selector.grid(row=3, column=3)
    ###
    tk.Label(
        canvas,
        text="Exhaustive search",
        justify=tk.RIGHT,
        anchor="e",
        font=wizard.style_font_bold_small,
    ).grid(row=4, column=0, sticky="ew")
    exhaustive_search = tk.IntVar(canvas)
    exhaustive_search.set(0)
    cb = tk.Checkbutton(canvas, onvalue=1, offvalue=0, variable=exhaustive_search)
    cb.grid(row=4, column=1)

    start_button = tk.Button(
        canvas,
        text="Start Optimizer for " + execution_configuration.value_dict["label"],
        command=lambda: __start_optimizer(
            execution_configuration,
            compile_command,
            doall_microbench_file,
            reduction_microbench_file,
            exhaustive_search,
            parent_frame,
        ),
    )
    start_button.grid(row=5, column=0)


def __start_optimizer(
    execution_configuration,
    compile_command,
    doall_microbench_file,
    reduction_microbench_file,
    exhaustive_search,
    parent_frame,
):
    arguments = {
        "--project": execution_configuration.value_dict["project_path"],
        "--detection-result-dump": get_path(
            execution_configuration.value_dict["working_copy_path"], "detection_result_dump.json"
        ),
        "--execute-created-models": False,
        "--clean-created-code": False,
        "--code-export-path": get_path(
            execution_configuration.value_dict["project_path"], ".discopop_optimizer/code_exports"
        ),
        "--dp-output-path": execution_configuration.value_dict["working_copy_path"],
        "--file-mapping": get_path(execution_configuration.value_dict["working_copy_path"], "FileMapping.txt"),
        "--executable-arguments": execution_configuration.value_dict["executable_arguments"],
        "--executable-name": execution_configuration.value_dict["executable_name"],
        "--linker-flags": execution_configuration.value_dict["linker_flags"],
        "--make-target": execution_configuration.value_dict["make_target"],
        "--make-flags": execution_configuration.value_dict["make_flags"],
        "--execution-repetitions": 1,
        "--execute-single-model": False,
        "--compile-command": compile_command.get(),
        "--execution-append-measurements": False,
        "--exhaustive-search": True if exhaustive_search.get() == 1 else False,
        "--headless-mode": False,
        "--doall-microbench-file": doall_microbench_file.get(),
        "--reduction-microbench-file": reduction_microbench_file.get(),
        "--dp-optimizer-path": get_path(execution_configuration.value_dict["project_path"], ".discopop_optimizer"),
    }

    # close elements on optimizer_frame
    for c in parent_frame.winfo_children():
        c.destroy()

    # start_optimizer(arguments, parent_frame=parent_frame)
    warnings.warn("TODO: BINDING TO OPTIMIZER CURRENTLY NOT IMPLEMENTED!")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Dict, List, Tuple, TypeVar, Union, overload

import numpy as np

from .Microbench import (
    Microbench,
    MicrobenchType,
    MicrobenchDimension,
    MicrobenchCoordinate,
)

T = TypeVar("T")


def __partition(pred, list: List[T]) -> Tuple[List[T], List[T]]:
    trues: List[T] = []
    falses: List[T] = []
    for item in list:
        if pred(item):
            trues.append(item)
        else:
            falses.append(item)
    return trues, falses


def __remove_outliers_using_iqr(values: List[float], iqr_factor=1.5) -> List[float]:
    values.sort()
    q1, q3 = np.percentile(values, [25, 75], method="linear")
    iqr = q3 - q1
    lower_bound = q1 - iqr_factor * iqr
    upper_bound = q3 + iqr_factor * iqr
    trimmed, outliers = __partition(lambda x: (x >= lower_bound and x <= upper_bound), values)
    return trimmed


@dataclass
class PureDataMicrobench(Microbench):
    measurements: Dict[
        MicrobenchType,
        Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],
    ]

    @overload
    def __getitem__(self, key: MicrobenchType) -> Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]]:
        ...

    @overload
    def __getitem__(self, key: Tuple[MicrobenchType, MicrobenchDimension]) -> Dict[MicrobenchCoordinate, List[float]]:
        ...

    @overload
    def __getitem__(self, key: Tuple[MicrobenchType, MicrobenchDimension, MicrobenchCoordinate]) -> List[float]:
        ...

    # allow to use this class like a dictionary
    def __getitem__(self, key):
        if isinstance(key, MicrobenchType):
            return self.measurements[key]
        elif (
            (isinstance(key, tuple) or isinstance(key, List))
            and len(key) == 2
            and isinstance(key[0], MicrobenchType)
            and isinstance(key[1], MicrobenchDimension)
        ):
            return self.measurements[key[0]][key[1]]
        elif (
            (isinstance(key, tuple) or isinstance(key, List))
            and len(key) == 3
            and isinstance(key[0], MicrobenchType)
            and isinstance(key[1], MicrobenchDimension)
            and isinstance(key[2], MicrobenchCoordinate)
        ):
            return self.measurements[key[0]][key[1]][key[2]]
        else:
            raise KeyError("Invalid key type:" + str(type(key)))

    # allow to use this class like a dictionary
    def __setitem__(self, key, value):
        if isinstance(key, MicrobenchType):
            if not isinstance(value, Dict):
                raise ValueError(
                    "Invalid value type: "
                    + str(type(value))
                    + " expected: Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]]"
                )
            self.measurements[key] = value
        elif (
            (isinstance(key, tuple) or isinstance(key, List))
            and len(key) == 2
            and isinstance(key[0], MicrobenchType)
            and isinstance(key[1], MicrobenchDimension)
        ):
            if not isinstance(value, Dict):
                raise ValueError(
                    "Invalid value type: " + str(type(value)) + " expected: Dict[MicrobenchCoordinate, List[float]]"
                )
            self.measurements[key[0]][key[1]] = value
        elif (
            (isinstance(key, tuple) or isinstance(key, List))
            and len(key) == 3
            and isinstance(key[0], MicrobenchType)
            and isinstance(key[1], MicrobenchDimension)
            and isinstance(key[2], MicrobenchCoordinate)
        ):
            if not isinstance(value, List):
                self.measurements[key[0]][key[1]][key[2]] = value
        else:
            raise KeyError("Invalid key type:" + str(type(key)))

    def removeOutliers(self):
        for type, dimMap in self.measurements.items():
            for dim, coordMap in dimMap.items():
                for coord, values in coordMap.items():
                    self.measurements[type][dim][coord] = __remove_outliers_using_iqr(values)

    def useMedian(self):
        for type, dimMap in self.measurements.items():
            for dim, coordMap in dimMap.items():
                for coord, values in coordMap.items():
                    self.measurements[type][dim][coord] = [np.median(values).item()]

    def useMean(self):
        for type, dimMap in self.measurements.items():
            for dim, coordMap in dimMap.items():
                for coord, values in coordMap.items():
                    self.measurements[type][dim][coord] = [np.mean(values).item()]

    def removeZeroParameters(self):
        for type, dimMap in self.measurements.items():
            for dim, coordMap in dimMap.items():
                for coord in list(coordMap.keys()):
                    if 0 in coord:
                        del self.measurements[type][dim][coord]

    def clampValues(
        self,
        dim: List[MicrobenchDimension] = [
            MicrobenchDimension.REFERENCE,
            MicrobenchDimension.TEST,
            MicrobenchDimension.OVERHEAD,
        ],
        min: float = float("-inf"),
        max: float = float("inf"),
    ):
        for type, dimMap in self.measurements.items():
            for d in dim:
                if d in dimMap:
                    for coord, values in dimMap[d].items():
                        self.measurements[type][d][coord] = [
                            v if v >= min and v <= max else (min if v < min else max) for v in values
                        ]

    def merge(self, other: PureDataMicrobench):
        for type, dimMap in other.measurements.items():
            if type not in self.measurements:
                self.measurements[type] = {}
            for dim, coordMap in dimMap.items():
                if dim not in self.measurements[type]:
                    self.measurements[type][dim] = {}
                for coord, values in coordMap.items():
                    if coord not in self.measurements[type][dim]:
                        self.measurements[type][dim][coord] = []
                    self.measurements[type][dim][coord].extend(values)
        # TODO check for duplicates? (but maybe we want them?)

    def mergeAll(self, others: List[PureDataMicrobench]):
        for other in others:
            self.merge(other)

    # inherited from Microbench
    def getMeasurements(
        self,
    ) -> Dict[MicrobenchType, Dict[MicrobenchDimension, Dict[MicrobenchCoordinate, List[float]]],]:
        return self.measurements

    # inherited from Microbench
    def toJSON(self) -> str:
        return json.dumps(
            {
                "measurements": {
                    type.value: {
                        dim.value: [
                            {"point": list(point), "values": values}
                            for point, values in self.measurements[type][dim].items()
                        ]
                        for dim in self.measurements[type].keys()
                        # for dim in MicrobenchDimension
                    }
                    for type in self.measurements.keys()
                    # for type in MicrobenchType
                },
                "parameters": ["Threads", "Workload", "Iterations"],
            }
        )

    # inherited from Microbench
    def evaluateInterpolation(
        self,
        benchType: MicrobenchType,
        benchDim: MicrobenchDimension,
        benchCoord: Union[MicrobenchCoordinate, Tuple[int, float, float]],
    ):
        raise TypeError("This class does not support interpolation.")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import json
import os

from filelock import FileLock  # type: ignore
from discopop_explorer.PEGraphX import LineID, Node, NodeID


class PatternBase(object):
    """Base class for pattern info"""

    pattern_id: int
    _node: Node
    node_id: NodeID
    start_line: LineID
    end_line: LineID

    def __init__(self, node: Node):
        # create a file lock to synchronize processes
        with FileLock(os.path.join(os.getcwd(), "next_free_pattern_id.txt.lock")):
            with open(os.path.join(os.getcwd(), "next_free_pattern_id.txt"), "r+") as f:
                lines = f.readlines()
                f.truncate(0)
                f.seek(0)
                if len(lines) == 0:
                    self.pattern_id = 0
                    f.write(str(0))
                else:
                    for line in lines:
                        line = line.replace("\n", "").replace("\x00", "")
                        self.pattern_id = int(line)
                        f.write(str(self.pattern_id + 1))
        self._node = node
        self.node_id = node.id
        self.start_line = node.start_position()
        self.end_line = node.end_position()

    def to_json(self):
        dic = self.__dict__
        keys = [k for k in dic.keys()]
        for key in keys:
            if key.startswith("_"):
                del dic[key]

        return json.dumps(dic, indent=2, default=lambda o: o.toJSON())  # , default=lambda o: "<not serializable>")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from json import JSONEncoder
from typing import Dict, Any
from discopop_explorer.pattern_detectors.PatternBase import PatternBase
from discopop_library.discopop_optimizer.classes.context.Update import Update
from discopop_library.discopop_optimizer.classes.types.DataAccessType import WriteDataAccess

from discopop_library.result_classes.DetectionResult import DetectionResult
from discopop_library.result_classes.PatternStorage import PatternStorage
from .PEGraphX import Node
from .pattern_detectors.PatternInfo import PatternInfo
from .pattern_detectors.pipeline_detector import PipelineStage
from .pattern_detectors.task_parallelism.classes import TPIType
from .variable import Variable


def filter_members(d: Dict[Any, Any]) -> Dict[Any, Any]:
    """Removes private and protected members (which starts with '_')

    :param d: member dictionary
    :return: member dictionary
    """
    keys = [k for k in d.keys()]
    for key in keys:
        if key.startswith("_"):
            del d[key]
    return d


class PatternBaseSerializer(JSONEncoder):
    """Json Encoder for Pattern Info"""

    def default(self, o):
        try:
            iterable = iter(o)
        except TypeError:
            pass
        else:
            return list(iterable)

        if isinstance(o, Variable):
            if o.operation is not None and o.operation != "":
                return f"{o.operation}:{o.name}"
            return o.name
        if isinstance(o, PatternBase):
            return filter_members(o.__dict__)
        if isinstance(o, DetectionResult):
            return filter_members(o.__dict__)
        if isinstance(o, PatternStorage):
            return filter_members(o.__dict__)
        if isinstance(o, PipelineStage):
            return filter_members(o.__dict__)
        if isinstance(o, Update):
            return filter_members(o.__dict__)
        if isinstance(o, WriteDataAccess):
            return filter_members(o.__dict__)
        if isinstance(o, Node):
            return o.id
        if isinstance(o, TPIType):
            return o.value

        # Let the base class default method raise the TypeError
        return JSONEncoder.default(self, o)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


from typing import List, Tuple, Dict, Set

from alive_progress import alive_bar  # type: ignore

from .PatternInfo import PatternInfo
from ..PEGraphX import (
    CUNode,
    LineID,
    LoopNode,
    NodeID,
    PEGraphX,
    Node,
    EdgeType,
    DepType,
    Dependency,
)
from ..utils import correlation_coefficient, classify_task_vars

__pipeline_threshold = 0.9


class PipelineStage(object):
    def __init__(self, pet: PEGraphX, node: Node, in_dep, out_dep):
        self.node = node.id
        self.startsAtLine = node.start_position()
        self.endsAtLine = node.end_position()

        fp, p, s, in_deps, out_deps, in_out_deps, r = classify_task_vars(pet, node, "Pipeline", in_dep, out_dep)

        self.first_private = fp
        self.private = p
        self.shared = s
        self.reduction = r
        self.in_deps = in_deps
        self.out_deps = out_deps
        self.in_out_deps = in_out_deps

    def __str__(self):
        return (
            f"\tNode: {self.node}\n"
            f"\tStart line: {self.startsAtLine}\n"
            f"\tEnd line: {self.endsAtLine}\n"
            f'\tpragma: "#pragma omp task"\n'
            f"\tfirst private: {[v.name for v in self.first_private]}\n"
            f"\tprivate: {[v.name for v in self.private]}\n"
            f"\tshared: {[v.name for v in self.shared]}\n"
            f"\treduction: {[v for v in self.reduction]}\n"
            f"\tInDeps: {[v.name for v in self.in_deps]}\n"
            f"\tOutDeps: {[v.name for v in self.out_deps]}\n"
            f"\tInOutDeps: {[v.name for v in self.in_out_deps]}"
        )


class PipelineInfo(PatternInfo):
    """Class, that contains pipeline detection result"""

    coefficient: float

    def __init__(self, pet: PEGraphX, node: Node):
        """
        :param pet: PET graph
        :param node: node, where pipeline was detected
        """
        PatternInfo.__init__(self, node)
        self._pet = pet
        self.coefficient = round(node.pipeline, 3)

        children_start_lines = [v.start_position() for v in pet.subtree_of_type(node, LoopNode)]

        self._stages = [
            pet.node_at(t)
            for s, t, d in pet.out_edges(node.id, [EdgeType.CHILD, EdgeType.CALLSNODE])
            if is_pipeline_subnode(node, pet.node_at(t), children_start_lines)
        ]

        self.stages = [self.__output_stage(s) for s in self._stages]

    def __in_dep(self, node: Node):
        raw: List[Tuple[NodeID, NodeID, Dependency]] = []
        for n in self._pet.subtree_of_type(node, CUNode):
            raw.extend((s, t, d) for s, t, d in self._pet.out_edges(n.id, EdgeType.DATA) if d.dtype == DepType.RAW)

        nodes_before = [node]
        for i in range(self._stages.index(node)):
            nodes_before.extend(self._pet.subtree_of_type(self._stages[i], CUNode))

        return [dep for dep in raw if dep[1] in [n.id for n in nodes_before]]

    def __out_dep(self, node: Node):
        raw: List[Tuple[NodeID, NodeID, Dependency]] = []
        for n in self._pet.subtree_of_type(node, CUNode):
            raw.extend((s, t, d) for s, t, d in self._pet.in_edges(n.id, EdgeType.DATA) if d.dtype == DepType.RAW)

        nodes_after = [node]
        for i in range(self._stages.index(node) + 1, len(self._stages)):
            nodes_after.extend(self._pet.subtree_of_type(self._stages[i], CUNode))

        return [dep for dep in raw if dep[0] in [n.id for n in nodes_after]]

    def __output_stage(self, node: Node) -> PipelineStage:
        in_d = self.__in_dep(node)
        out_d = self.__out_dep(node)

        return PipelineStage(self._pet, node, in_d, out_d)

    def __str__(self):
        s = "\n\n".join([str(s) for s in self.stages])
        return (
            f"Pipeline at: {self.node_id}\n"
            # f"Coefficient: {round(self.coefficient, 3)}\n"
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            f"Stages:\n{s}"
        )


def is_pipeline_subnode(root: Node, current: Node, children_start_lines: List[LineID]) -> bool:
    """Checks if node is a valid subnode for pipeline

    :param root: root node
    :param current: current node
    :param children_start_lines: start lines of children loops
    :return: true if valid
    """
    r_start = root.start_position()
    r_end = root.end_position()
    c_start = current.start_position()
    c_end = current.end_position()
    return not (
        c_start == r_start
        and c_end == r_start
        or c_start == r_end
        and c_end == r_end
        or c_start == c_end
        and c_start in children_start_lines
    )


global_pet = None


def run_detection(pet: PEGraphX) -> List[PipelineInfo]:
    """Search for pipeline pattern on all the loops in the graph
    except for doall loops

    :param pet: PET graph
    :return: List of detected pattern info
    """
    import tqdm  # type: ignore
    from multiprocessing import Pool

    global global_pet
    global_pet = pet

    result: List[PipelineInfo] = []
    children_cache: Dict[Node, List[Node]] = dict()
    dependency_cache: Dict[Tuple[Node, Node], Set[Node]] = dict()
    nodes = pet.all_nodes(LoopNode)

    param_list = [(node) for node in nodes]
    with Pool(initializer=__initialize_worker, initargs=(pet,)) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__check_node, param_list), total=len(param_list)))
    for local_result in tmp_result:
        result += local_result
    print("GLOBAL RES: ", result)

    for pattern in result:
        pattern.get_workload(pet)

    return result


def __initialize_worker(pet):
    global global_pet
    global_pet = pet


def __check_node(param_tuple):
    global global_pet
    local_result = []

    node = param_tuple

    node.pipeline = __detect_pipeline(global_pet, node)
    if node.pipeline > __pipeline_threshold:
        local_result.append(PipelineInfo(global_pet, node))

    return local_result


def __detect_pipeline(pet: PEGraphX, root: Node, children_cache=None, dep_cache=None) -> float:
    """Calculate pipeline value for node

    :param pet: PET graph
    :param root: current node
    :param children_cache: used to cache intermediate children
    :param dep_cache: used to cache intermediate dependencies
    :return: Pipeline scalar value
    """

    children_start_lines = [v.start_position() for v in pet.subtree_of_type(root, LoopNode)]

    loop_subnodes = [
        pet.node_at(t)
        for s, t, d in pet.out_edges(root.id, [EdgeType.CHILD, EdgeType.CALLSNODE])
        if is_pipeline_subnode(root, pet.node_at(t), children_start_lines)
    ]

    # No chain of stages found
    if len(loop_subnodes) < 2:
        return 0

    graph_vector = []
    for i in range(0, len(loop_subnodes) - 1):
        graph_vector.append(1.0 if pet.depends_ignore_readonly(loop_subnodes[i + 1], loop_subnodes[i], root) else 0.0)

    pipeline_vector = []
    for i in range(0, len(loop_subnodes) - 1):
        pipeline_vector.append(1.0)

    min_weight = 1.0
    for i in range(0, len(loop_subnodes) - 1):
        for j in range(i + 1, len(loop_subnodes)):
            if pet.depends_ignore_readonly(loop_subnodes[i], loop_subnodes[j], root):
                # TODO whose corresponding entry in the graph matrix is nonzero?
                node_weight = 1 - (j - i) / (len(loop_subnodes) - 1)
                if min_weight > node_weight > 0:
                    min_weight = node_weight

    if min_weight == 1.0:
        graph_vector.append(0.0)
        pipeline_vector.append(0)
    else:
        graph_vector.append(1.0)
        pipeline_vector.append(min_weight)

    return correlation_coefficient(graph_vector, pipeline_vector)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import sys
from pathlib import Path

from setuptools import setup, find_packages  # type:ignore

from discopop_library.global_data.version.utils import get_version

os.chdir(Path(__file__).parent)

if sys.version_info < (3, 6):
    raise SystemExit("Discopop explorer requires Python >= 3.6.")

setup(
    name="discopop",
    version=get_version(),
    packages=find_packages(),
    url="https://www.discopop.tu-darmstadt.de/",
    author="TU Darmstadt and Iowa State University",
    author_email="discopop@lists.parallel.informatik.tu-darmstadt.de",
    description="DiscoPoP is a tool that helps software developers parallelize their "
    "programs with threads. It discovers potential parallelism in a "
    "sequential program and makes recommendations on how to exploit it.",
    # long_description=open(SRC / "README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "contextlib2>=0.5.5",
        "argparse",
        "extrap",
        "jsonpickle",
        "jsons",
        "lxml>=4.3.3",
        "matplotlib",
        "networkx",
        "numpy>=1.16.3",
        "pluginbase>=1.0.0",
        "pstats2",
        "schema>=0.7.0",
        "sympy",
        "sympy_plot_backends",
        "alive_progress",
        "filelock",
    ],
    extras_require={
        "dev": ["mypy", "black", "data-science-types", "pre-commit"],
        "ci": ["mypy", "black", "data-science-types"],
    },
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "discopop_explorer=discopop_explorer.__main__:main",
            "DEPRECATED_discopop_profiler=DEPRECATED_discopop_profiler.__main__:main",
            "DEPRECATED_discopop_wizard=DEPRECATED_discopop_wizard.__main__:main",
            "discopop_optimizer=discopop_library.discopop_optimizer.__main__:main",
            "discopop_patch_generator=discopop_library.PatchGenerator.__main__:main",
            "discopop_patch_applicator=discopop_library.PatchApplicator.__main__:main",
            "discopop_config_provider=discopop_library.ConfigProvider.__main__:main",
        ]
    },
    zip_safe=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
    ],
    license_files=["LICENSE"],
    include_package_data=True,
)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import copy

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode


class ContextSnapshot(ContextNode):
    def __init__(self, node_id: int, experiment):
        super().__init__(node_id, experiment)

    def get_plot_label(self) -> str:
        return str(self.node_id) + "\nCTX\nsnapshot"

    def get_modified_context(
        self, node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject
    ) -> ContextObject:
        context.snapshot_stack.append(
            (
                context.seen_writes_by_device,
                context.last_visited_node_id,
                context.last_seen_device_ids,
                context.necessary_updates,
            )
        )
        context.seen_writes_by_device = dict()
        context.last_seen_device_ids = [context.last_seen_device_ids[-1]]
        context.necessary_updates = set()

        context.save_stack.append([])
        return context



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from argparse import ArgumentParser

from discopop_library.PatchApplicator.PatchApplicatorArguments import PatchApplicatorArguments
from discopop_library.PatchApplicator.patch_applicator import run


def parse_args() -> PatchApplicatorArguments:
    """Parse the arguments passed to the discopop_explorer"""
    parser = ArgumentParser(description="DiscoPoP Patch Generator")
    # all flags that are not considered stable should be added to the experimental_parser
    experimental_parser = parser.add_argument_group(
        "EXPERIMENTAL",
        "Arguments for the task pattern detector and other experimental features. These flags are considered EXPERIMENTAL and they may or may not be removed or changed in the near future.",
    )

    # fmt: off
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="Enable verbose output.")
    parser.add_argument('-a', '--apply', nargs='+', default=[], help="Apply the parallelization suggestions with the "
                                                                     "given ids.")
    parser.add_argument('-r', '--rollback', nargs='+', default=[], help="Roll back the application of the "
                                                                        "parallelization suggestions with the given "
                                                                        "ids.")
    parser.add_argument('-C', '--clear', action="store_true", help="Reset the code to it's original state. Preserves "
                                                                   "the list of applied suggestion for loading."
                                                                   "Old saves will be overwritten!")
    # todo: in the long run, saving configurations to different locations could be easily implemented.

    parser.add_argument('-L', '--load', action="store_true", help="Load a previous state after clearing.")
    parser.add_argument('-l', '--list', action="store_true", help="Show the list of applied suggestions."
                                                                  "If set, nothing else will be done.")
    # EXPERIMENTAL FLAGS:
    # fmt: on

    arguments = parser.parse_args()

    return PatchApplicatorArguments(
        verbose=arguments.verbose,
        apply=arguments.apply,
        rollback=arguments.rollback,
        clear=arguments.clear,
        load=arguments.load,
        list=arguments.list,
    )


def main() -> int:
    """Return values:"
    "0: Applied successfully"
    "1: Nothing applied"
    "2: Some changes applied successfully
    """
    retval = 0
    arguments = parse_args()
    retval = run(arguments)
    return retval


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter as tk
from tkinter import ttk


class ScrollableFrameWidget(object):
    container: ttk.Frame
    canvas: tk.Canvas
    scrollbar: tk.Scrollbar
    scrollable_frame: tk.Frame  # important

    def __init__(self, parent_frame):
        self.container = ttk.Frame(parent_frame)
        self.canvas = tk.Canvas(self.container)
        self.scrollbar = ttk.Scrollbar(self.container, orient="vertical", command=self.canvas.yview)
        self.scrollable_frame = ttk.Frame(self.canvas)

        # configure weights
        self.container.rowconfigure(0, weight=1)
        self.container.columnconfigure(0, weight=1)
        self.canvas.rowconfigure(0, weight=1)
        self.canvas.columnconfigure(0, weight=1)
        self.scrollable_frame.rowconfigure(0, weight=1)
        self.scrollable_frame.columnconfigure(0, weight=1)

        self.scrollable_frame.bind("<Configure>", lambda e: self.canvas.configure(scrollregion=self.canvas.bbox("all")))

        self.canvas.create_window((0, 0), window=self.scrollable_frame, anchor="nw")

        self.canvas.configure(yscrollcommand=self.scrollbar.set)

    def finalize(self, row_count: int, row: int = 0, col: int = 0, rowspan: int = 1, columnspan: int = 1):
        if rowspan < 1:
            rowspan = 1
        if columnspan < 1:
            columnspan = 1
        self.container.grid(row=row, column=col, columnspan=columnspan, rowspan=rowspan, sticky=tk.NSEW)
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)
        self.scrollbar.grid(row=0, rowspan=max(row_count, 1), column=1, sticky=tk.NS)

    def get_scrollable_frame(self) -> tk.Frame:
        return self.scrollable_frame



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


def get_unique_pattern_id() -> int:
    with open("next_free_pattern_id.txt", "r") as f:
        pattern_id = int(f.read())
    with open("next_free_pattern_id.txt", "w") as f:
        f.write(str(pattern_id + 1))
    return pattern_id



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from sympy import Symbol  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.edges.GenericEdge import GenericEdge


class DataFlowEdge(GenericEdge):
    """Used to represent data flow in the Graph. Used to determine cutoff points for local optimization"""

    def get_cost_model(self) -> CostModel:
        raise ValueError("The cost of a DataFlowEdge is not defined and may never be used!")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import sys

from alive_progress import alive_bar  # type: ignore

from discopop_explorer.pattern_detectors.task_parallelism.task_parallelism_detector import (
    build_preprocessed_graph_and_run_detection as detect_tp,
)
from discopop_library.discopop_optimizer.OptimizationGraph import OptimizationGraph
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.system.System import System
from discopop_library.discopop_optimizer.scheduling.workload_delta import (
    get_workload_delta_for_cu_node,
)
from discopop_library.discopop_optimizer.utilities.MOGUtilities import get_nodes_from_cu_id
from discopop_library.result_classes.DetectionResult import DetectionResult
from .PEGraphX import DummyNode, LoopNode, PEGraphX, EdgeType
from .pattern_detectors.do_all_detector import run_detection as detect_do_all
from .pattern_detectors.geometric_decomposition_detector import run_detection as detect_gd
from .pattern_detectors.pipeline_detector import run_detection as detect_pipeline
from .pattern_detectors.reduction_detector import run_detection as detect_reduction
from .pattern_detectors.simple_gpu_patterns.gpu_pattern_detector import run_detection as detect_gpu


class PatternDetectorX(object):
    pet: PEGraphX

    def __init__(self, pet_graph: PEGraphX) -> None:
        """This class runs detection algorithms on CU graph

        :param pet_graph: CU graph
        """
        self.pet = pet_graph

    def __merge(self, loop_type: bool, remove_dummies: bool):
        """Removes dummy nodes

        :param loop_type: loops only
        :param remove_dummies: remove dummy nodes
        """
        dummies_to_remove = set()
        for node in self.pet.all_nodes():
            if not loop_type or isinstance(node, LoopNode):
                if remove_dummies and isinstance(node, DummyNode):
                    continue
                for s, t, e in self.pet.out_edges(node.id, [EdgeType.CHILD, EdgeType.CALLSNODE]):
                    if remove_dummies and isinstance(self.pet.node_at(t), DummyNode):
                        dummies_to_remove.add(t)

        for n in dummies_to_remove:
            self.pet.g.remove_node(n)

    def detect_patterns(
        self,
        project_path,
        cu_dict,
        dependencies,
        loop_data,
        reduction_vars,
        file_mapping,
        cu_inst_result_file,
        llvm_cxxfilt_path,
        discopop_build_path,
        enable_patterns,
        enable_task_pattern,
        enable_detection_of_scheduling_clauses,
    ):
        """Runs pattern discovery on the CU graph"""
        self.__merge(False, True)
        self.pet.map_static_and_dynamic_dependencies()
        self.pet.calculateFunctionMetadata()
        self.pet.calculateLoopMetadata()
        res = DetectionResult(self.pet)

        # reduction before doall!

        if "*" in enable_patterns or "reduction" in enable_patterns:
            print("REDUCTIONS...")
            res.patterns.reduction = detect_reduction(self.pet)
            print("\tDONE.")
        if "*" in enable_patterns or "doall" in enable_patterns:
            print("DOALL...")
            res.patterns.do_all = detect_do_all(self.pet)
            print("\tDONE.")
        if "*" in enable_patterns or "pipeline" in enable_patterns:
            print("PIPELINE...")
            res.patterns.pipeline = detect_pipeline(self.pet)
            print("\tDONE.")
        if "*" in enable_patterns or "geodec" in enable_patterns:
            print("GEO. DEC...")
            res.patterns.geometric_decomposition = detect_gd(self.pet)
            print("\tDONE.")

        # check if task pattern should be enabled
        if enable_task_pattern:
            res.patterns.task = detect_tp(
                cu_dict,
                dependencies,
                reduction_vars,
                file_mapping,
                cu_inst_result_file,
                llvm_cxxfilt_path,
                discopop_build_path,
            )

        # detect GPU patterns based on previously identified patterns
        if "*" in enable_patterns or "simplegpu" in enable_patterns:
            print("SIMPLE GPU...")
            res.patterns.simple_gpu = detect_gpu(self.pet, res, project_path)
            print("\tDONE.")

        # detect combined GPU patterns
        # disabled currently due to high additional overhead.
        # will be moved and calculated based on the optimization graph
        # res.combined_gpu = detect_combined_gpu(self.pet, res, project_folder_path)

        #        if enable_detection_of_scheduling_clauses:
        # identify scheduling clauses
        #            return self.__identify_scheduling_clauses(res, project_path, file_mapping)
        return res


# todo: re-enable identification of scheduling clauses
#   def __identify_scheduling_clauses(
#       self,
#       res: DetectionResult,
#       project_folder_path: str,
#       file_mapping_path: str,
#   ) -> DetectionResult:
#       """Identifies scheduling clauses for suggestions and returns the updated DetectionResult"""
#       # construct optimization graph (basically an acyclic representation of the PET)
#       system = System()
#       discopop_output_path = project_folder_path
#       discopop_optimizer_path = "INVALID_DUMMY"
#       code_export_path = "INVALID_DUMMY"
#       arguments_1 = {"--compile-command": "make"}
#       experiment = Experiment(
#           project_folder_path,
#           discopop_output_path,
#           discopop_optimizer_path,
#           code_export_path,
#           file_mapping_path,
#           system,
#           res,
#           arguments_1,
#       )
#       arguments_2 = {"--exhaustive-search": False, "--headless-mode": True}
#       optimization_graph = OptimizationGraph(project_folder_path, experiment, arguments_2, None, False)
#
#        for do_all_suggestion in res.do_all:
#            for node_id in get_nodes_from_cu_id(experiment.optimization_graph, do_all_suggestion.node_id):
#                workload_delta, min_workload, max_workload = get_workload_delta_for_cu_node(experiment, node_id)
#                print(
#                    "DOALL @ ",
#                    do_all_suggestion.node_id,
#                    " -> ",
#                    "node_id: ",
#                    node_id,
#                    " --> Delta WL: ",
#                    workload_delta,
#                    " (",
#                    min_workload,
#                    "/",
#                    max_workload,
#                    ")",
#                    file=sys.stderr,
#                )
#                # todo
#                #  very naive and non-robust approach, needs improvement in the future
#                #  reflects the behavior as described in https://dl.acm.org/doi/pdf/10.1145/3330345.3330375
#                if workload_delta != 0:
#                    do_all_suggestion.scheduling_clause = "dynamic"
#
#        return res



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.utilities.MOGUtilities import get_parents
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern


def fix_loop_initialization_updates(
    experiment: Experiment, best_configuration: OptimizerOutputPattern, arguments: OptimizerArguments
) -> OptimizerOutputPattern:
    """Move updates to initialize device loops before the loop"""
    for update in best_configuration.data_movement:
        if update.source_node_id == update.target_node_id:
            continue
        source_parents = [
            id
            for id in get_parents(experiment.optimization_graph, update.source_node_id)
            if id in best_configuration.decisions
        ]
        target_parents = [
            id
            for id in get_parents(experiment.optimization_graph, update.target_node_id)
            if id in best_configuration.decisions
        ]

        if len(source_parents) != 0 and len(target_parents) != 0:
            source_parent = source_parents[0]
            target_parent = target_parents[0]

            if source_parent == target_parent and type(data_at(experiment.optimization_graph, source_parent)) == Loop:
                # set the loop node as the new update source and target
                update.source_node_id = source_parent
                update.source_cu_id = data_at(experiment.optimization_graph, source_parent).original_cu_id
                update.target_node_id = source_parent
                update.target_cu_id = update.source_cu_id
                continue

        if len(target_parents) != 0:
            # move updates originating from the loop node to the inside before the loop
            if (
                type(data_at(experiment.optimization_graph, update.source_node_id)) == Loop
                and update.source_node_id in target_parents
            ):
                update.target_node_id = update.source_node_id
                update.target_cu_id = data_at(experiment.optimization_graph, update.source_node_id).original_cu_id

    if arguments.verbose:
        print("Fixed loop initialization updates")
        for update in best_configuration.data_movement:
            print("# ", update)
        print()

    return best_configuration



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
from multiprocessing import Pool
import os
from typing import Dict, List, Optional, Set, Tuple, cast

from sympy import Expr
import tqdm  # type: ignore
from discopop_explorer.PEGraphX import NodeID  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.OptimizerArguments import OptimizerArguments
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.optimization.evaluate import evaluate_configuration
from itertools import product
from discopop_library.discopop_optimizer.optimization.validation import check_configuration_validity

from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from discopop_library.result_classes.OptimizerOutputPattern import OptimizerOutputPattern


global_experiment = None
global_arguments = None


def evaluate_all_decision_combinations(
    experiment: Experiment,
    available_decisions: Dict[FunctionRoot, List[List[int]]],
    arguments: OptimizerArguments,
    optimizer_dir: str,
) -> Optional[OptimizerOutputPattern]:
    """Create and evaluate every possible combination of decisions."""
    global global_experiment
    global global_arguments
    global_experiment = experiment
    global_arguments = arguments

    costs_dict: Dict[Tuple[int, ...], Expr] = dict()
    contexts_dict: Dict[Tuple[int, ...], ContextObject] = dict()

    combinations_by_function: Dict[FunctionRoot, List[Tuple[int, ...]]] = dict()
    for function in available_decisions:
        combinations_by_function[function] = []
        for cmb in product(*available_decisions[function]):
            combinations_by_function[function].append(cmb)
    # get list of all combinations
    combinations: List[List[int]] = []
    for c in product(*combinations_by_function.values()):
        combination_list: List[int] = []
        for function_decisions in c:
            for entry in function_decisions:
                combination_list.append(entry)
        # remove invalid combinations
        if check_configuration_validity(experiment, arguments, combination_list):
            combinations.append(combination_list)

    # evaluate each combination in parallel
    print("# Parallel calculation of costs of all decision combinations...")
    param_list = [(combination_list) for combination_list in combinations]
    with Pool(
        initializer=__initialize_worker,
        initargs=(
            experiment,
            arguments,
        ),
    ) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__evaluate_configuration, param_list), total=len(param_list)))

    #    tmp_result = []
    #    for p in param_list:
    #        tmp_result.append(__evaluate_configuration(p))

    for local_result in tmp_result:
        # result += local_result
        if local_result is not None:
            costs_dict[local_result[0]] = local_result[1]
            contexts_dict[local_result[0]] = local_result[2]

    if arguments.verbose:
        print()

        # print the sorted result for improved readability
        print("# Sorted and simplified costs of all combinations using PATH NODE IDS")
        for combination_tuple in sorted(costs_dict.keys(), key=lambda x: costs_dict[x], reverse=True):
            print(
                "#",
                combination_tuple,
                " = ",
                str(costs_dict[combination_tuple].evalf()),
            )
        print("# Sorted and simplified costs of all combinations using PATH NODE IDS")
        print()

        # print the sorted result list using parallel pattern ids for improved interpretability
        print("# Sorted and simplified costs of all combinations using PARALLEL PATTERN IDS")
        for combination_tuple in sorted(costs_dict.keys(), key=lambda x: costs_dict[x], reverse=True):
            new_key = []
            for node_id in combination_tuple:
                # find pattern id
                for pattern_id in experiment.suggestion_to_node_ids_dict:
                    if node_id in experiment.suggestion_to_node_ids_dict[pattern_id]:
                        new_key.append(
                            str(pattern_id) + "@" + str(data_at(experiment.optimization_graph, node_id).device_id)
                        )
            if len(new_key) == 0:
                print("EMPTY: ", combination_tuple)
                print(experiment.suggestion_to_node_ids_dict)

            print(
                "#",
                new_key,
                " = ",
                str(costs_dict[combination_tuple].evalf()),
            )
        print("# Sorted and simplified costs of all combinations using PARALLEL PATTERN IDS")
        print()

    return __dump_result_to_file_using_pattern_ids(experiment, optimizer_dir, costs_dict, contexts_dict, arguments)


def __initialize_worker(
    experiment: Experiment,
    arguments: OptimizerArguments,
):
    global global_experiment
    global global_arguments
    global_experiment = experiment
    global_arguments = arguments


def __evaluate_configuration(param_tuple):
    global global_experiment
    global global_arguments
    decisions = param_tuple
    return evaluate_configuration(global_experiment, decisions, global_arguments)


def __dump_result_to_file_using_pattern_ids(
    experiment: Experiment,
    optimizer_dir: str,
    costs_dict: Dict[Tuple[int, ...], Expr],
    contexts_dict: Dict[Tuple[int, ...], ContextObject],
    arguments: OptimizerArguments,
) -> Optional[OptimizerOutputPattern]:
    # replace keys to allow dumping
    dumpable_dict = dict()
    for key in costs_dict:
        new_key = []
        for entry in key:
            # find pattern id
            for pattern_id in experiment.suggestion_to_node_ids_dict:
                if entry in experiment.suggestion_to_node_ids_dict[pattern_id]:
                    new_key.append(str(pattern_id) + "@" + str(data_at(experiment.optimization_graph, entry).device_id))
        dumpable_dict[str(new_key)] = str(int(float(str(costs_dict[key].evalf()))))

    dump_path: str = os.path.join(optimizer_dir, "exhaustive_results.json")
    with open(dump_path, "w") as fp:
        json.dump(dumpable_dict, fp)

    # dump the best option
    for combination_tuple in sorted(costs_dict.keys(), key=lambda x: costs_dict[x]):
        new_key_2 = []
        best_configuration = None
        # collect applied suggestions
        for node_id in combination_tuple:
            # find pattern id
            for pattern_id in experiment.suggestion_to_node_ids_dict:
                if node_id in experiment.suggestion_to_node_ids_dict[pattern_id]:
                    new_key_2.append(
                        str(pattern_id) + "@" + str(data_at(experiment.optimization_graph, node_id).device_id)
                    )
                    device_id = data_at(experiment.optimization_graph, node_id).device_id
                    if best_configuration is None:
                        best_configuration = OptimizerOutputPattern(
                            experiment.detection_result.pet.node_at(
                                cast(NodeID, data_at(experiment.optimization_graph, node_id).original_cu_id)
                            ),
                            [],
                            experiment.get_system().get_host_device_id(),
                        )
                    best_configuration.add_pattern(
                        pattern_id, device_id, experiment.get_system().get_device(device_id).get_device_type()
                    )
        if best_configuration is None:
            return None
        # collect data movement information
        for update in contexts_dict[combination_tuple].necessary_updates:
            best_configuration.add_data_movement(update)
        # export results to file
        best_option_id_path: str = os.path.join(optimizer_dir, "exhaustive_pattern_id.txt")
        with open(best_option_id_path, "w+") as f:
            f.write(str(best_configuration.pattern_id))

        return best_configuration
    raise ValueError("No configuration found!")



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import json
import os.path
from typing import Dict, List


def read_patterns_from_json_to_json(json_path: str, relevant_patterns: List[str]) -> Dict[str, List[str]]:
    """relevant_patterns specifies the names of patterns which shall be returned.
    An empty list acts as a wildcard."""
    pattern_json_strings_by_type: Dict[str, List[str]] = dict()

    if not os.path.exists(json_path):
        raise ValueError("Path does not exist!", json_path)

    with open(json_path, "r") as f:
        data = json.load(f)
        for key in data["patterns"]:
            if len(relevant_patterns) > 0 and key not in relevant_patterns:
                continue
            if key not in pattern_json_strings_by_type:
                pattern_json_strings_by_type[key] = []
            for entry in data["patterns"][key]:
                pattern_json_strings_by_type[key].append(json.dumps(entry))

    return pattern_json_strings_by_type



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import networkx as nx  # type: ignore
from sympy import Integer

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload


class ContextNode(Workload):
    def __init__(self, node_id: int, experiment):
        super().__init__(
            node_id,
            experiment,
            cu_id=None,
            sequential_workload=Integer(0),
            parallelizable_workload=Integer(0),
        )

    def get_plot_label(self) -> str:
        return str(self.node_id) + "\nCTX"

    def get_modified_context(
        self, node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject
    ) -> ContextObject:
        return context



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from tkinter import *
from typing import Optional

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.gui.widgets.ScrollableFrame import ScrollableFrameWidget
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at
from DEPRECATED_discopop_wizard.screens.widgets.ScrollableText import ScrollableTextWidget


def display_choices_for_model(graph: nx.DiGraph, model: CostModel, window_title: Optional[str] = None):
    root = Tk()
    if window_title is not None:
        root.configure()
        root.title(window_title)
    # configure window size
    root.geometry("1000x600")
    # configure weights
    root.rowconfigure(0, weight=1)
    root.columnconfigure(0, weight=1)

    # create scrollable frame
    scrollable_frame_widget = ScrollableFrameWidget(root)
    scrollable_frame = scrollable_frame_widget.get_scrollable_frame()
    # configure weights
    scrollable_frame.rowconfigure(0, weight=1)
    scrollable_frame.rowconfigure(1, weight=1)
    scrollable_frame.rowconfigure(2, weight=1)
    scrollable_frame.columnconfigure(0, weight=1)
    scrollable_frame.columnconfigure(1, weight=1)
    scrollable_frame.columnconfigure(2, weight=1)

    rows = []
    # print mathematical function for model
    fn_label = Entry(scrollable_frame, relief=RIDGE, width=len("Function:"))
    fn_label.grid(row=0, column=0, sticky=NSEW)
    fn_label.insert(END, "Function:")
    fn_label.configure(state=DISABLED, disabledforeground="black")
    fn_function_frame = Frame(scrollable_frame)
    fn_function_frame.grid(row=0, column=1, columnspan=2, sticky=NSEW)
    fn_function = ScrollableTextWidget(fn_function_frame)
    fn_function.set_text(str(model.parallelizable_costs + model.sequential_costs))
    fn_function.text_container.config(height=3)

    column_headers = ["Decision", "Device", "Details"]
    # set column headers
    header_cols = []
    for col_idx, header in enumerate(column_headers):
        e = Entry(scrollable_frame, relief=RIDGE, width=len(header))
        e.grid(row=1, column=col_idx, sticky=NSEW)
        e.insert(END, header)
        e.configure(state=DISABLED, disabledforeground="black")
        header_cols.append(e)
    rows.append(header_cols)

    for row_idx, decision in enumerate(model.path_decisions):
        row_idx = row_idx + 2  # account for mathematical function and header row
        # add decision id
        e = Entry(scrollable_frame, relief=RIDGE)
        e.grid(row=row_idx, column=0, sticky=NSEW)
        e.insert(END, str(decision))
        e.configure(state=DISABLED, disabledforeground="black")

        # add device id
        i = Entry(scrollable_frame, relief=RIDGE)
        i.grid(row=row_idx, column=1, sticky=NSEW)
        i.insert(END, str(data_at(graph, decision).device_id))
        i.configure(state=DISABLED, disabledforeground="black")

        # add decision details
        d = Frame(scrollable_frame)
        d.grid(row=row_idx, column=2, sticky=NSEW)
        stw = ScrollableTextWidget(d)
        stw.set_text(str(data_at(graph, decision).suggestion))
        stw.text_container.config(height=10)

    scrollable_frame_widget.finalize(len(model.path_decisions))

    pass



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from enum import IntEnum


class PragmaPosition(IntEnum):
    BEFORE_START = 0
    AFTER_START = 1
    BEFORE_END = 2
    AFTER_END = 3


class OmpConstructPositioning(IntEnum):
    BEFORE_LINE = 0
    AFTER_LINE = 1


class PragmaType(IntEnum):
    PRAGMA = 1
    REGION = 2



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import cast

from sympy import Integer, Symbol

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject


def get_transfer_costs(context: ContextObject, environment: Experiment) -> CostModel:
    """Calculates the amount of data transferred between devices as specified by self.necessary_updates and
    calculates an estimation for the added transfer costs under the assumption,
    that no transfers happen concurrently and every transfer is executed in a blocking, synchronous manner.
    """
    total_transfer_costs = Integer(0)
    symbolic_memory_region_sizes = True
    symbol_value_suggestions = dict()
    for update in context.necessary_updates:
        # add static costs incurred by the transfer initialization
        system = environment.get_system()
        source_device = system.get_device(cast(int, update.source_device_id))
        target_device = system.get_device(cast(int, update.target_device_id))
        initialization_costs = system.get_network().get_transfer_initialization_costs(source_device, target_device)

        total_transfer_costs += initialization_costs

        # add costs incurred by the transfer itself
        transfer_speed = system.get_network().get_transfer_speed(source_device, target_device)

        # value suggestion used for symbolic values
        transfer_size, value_suggestion = environment.get_memory_region_size(
            update.write_data_access.memory_region,
            use_symbolic_value=symbolic_memory_region_sizes,
        )
        # save suggested memory region size from Environment
        if symbolic_memory_region_sizes:
            symbol_value_suggestions[cast(Symbol, transfer_size)] = value_suggestion

        # transfer_costs [us] = [1000000 * s]
        # transfer_size [B]
        # transfer_speed [MB/s] = [1000000 * B/s]
        # * 1000000 gets removed due to mathematics
        transfer_costs = transfer_size / transfer_speed

        total_transfer_costs += transfer_costs
    if symbolic_memory_region_sizes:
        return CostModel(Integer(0), total_transfer_costs, symbol_value_suggestions=symbol_value_suggestions)
    else:
        return CostModel(Integer(0), total_transfer_costs)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from discopop_library.discopop_optimizer.classes.edges.GenericEdge import GenericEdge


class ChildEdge(GenericEdge):
    pass



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from discopop_library.discopop_optimizer.classes.system.devices.Device import Device
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum


class CPU(Device):
    def __init__(
        self,
        frequency,
        thread_count,
        openmp_device_id: int,
        device_specific_compiler_flags: str,
    ):
        super().__init__(frequency, thread_count, openmp_device_id, device_specific_compiler_flags)

    def get_device_type(self) -> DeviceTypeEnum:
        return DeviceTypeEnum.CPU



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os.path
from pathlib import Path
from typing import Dict, Optional
import warnings


def load_file_mapping(fmap_path: str) -> Dict[int, Path]:
    if not os.path.exists(fmap_path):
        raise ValueError("Path does not exist! ", fmap_path)

    file_mapping: Dict[int, Path] = dict()

    with open(fmap_path, "r") as f:
        for line in f.readlines():
            line = line.replace("\n", "")
            split_line = line.split("\t")
            file_id = int(split_line[0])
            file_path = split_line[1]
            if os.path.exists(file_path):
                file_mapping[file_id] = Path(file_path)
            else:
                warnings.warn("Loading filemapping: File does not exist! " + str(file_path))
    return file_mapping


def get_path(base_path: str, file_name: str) -> str:
    """Combines path and filename if it is not absolute

    :param base_path: path
    :param file_name: file name
    :return: path to file
    """
    return file_name if os.path.isabs(file_name) else os.path.join(base_path, file_name)


def get_path_or_none(base_path: str, file_name: Optional[str]) -> Optional[str]:
    """same as get_path but returns None if file_name is None"""
    if file_name is None:
        return None
    return get_path(base_path, file_name)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, List, Optional, Tuple, cast

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_requirements,
    get_successors,
    get_children,
    show,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


def calculate_data_transfers(
    graph: nx.DiGraph, function_performance_models: Dict[FunctionRoot, List[CostModel]], experiment
) -> Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]]:
    """Calculate data transfers for each performance model and append the respective ContextObject to the result."""
    result_dict: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]] = dict()
    for function in function_performance_models:
        result_dict[function] = []
        for model in function_performance_models[function]:
            # create a ContextObject for the current path
            context = ContextObject(function.node_id, [function.device_id])
            context = get_path_context_iterative(function.node_id, graph, model, context, experiment)
            result_dict[function].append((model, context))
    return result_dict


def get_path_context_iterative(
    root_node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject, experiment
) -> ContextObject:
    """passes the context Object along the path and returns the context once the end has been reached"""
    node_id = None
    next_node_id: Optional[int] = root_node_id

    while next_node_id is not None:
        node_id = next_node_id

        # push device id to stack if necessary
        node_data = data_at(graph, node_id)
        if node_data.device_id is not None:
            context.last_seen_device_ids.append(node_data.device_id)

        # calculate context modifications for the current node
        context = __check_current_node(node_id, graph, model, context, experiment)

        # set last_visited_device id
        if node_data.device_id is not None:
            context.last_visited_device_id = node_data.device_id
        else:
            context.last_visited_device_id = context.last_seen_device_ids[-1]

        # calculate context modifications for the children of the current node
        context = __check_children(node_id, graph, model, context, experiment)

        # pop device id from stack if necessary
        if node_data.device_id is not None:
            # condition to prevent issues due to added dummy nodes
            if len(context.last_seen_device_ids) > 1:
                context.last_seen_device_ids.pop()

        # set last_visited_node_id to the original node_id,
        # since the calculation continues from node_id after the children have been visited
        context.last_visited_node_id = node_id

        # pass context to the right successor of the current node
        # At most a single node can be a successor, since the given model represents a single path through the graph.
        successors = get_successors(graph, node_id)
        if len(successors) == 0:
            # stop the loop
            next_node_id = None
        elif len(successors) == 1:
            # pass context to the single successor
            next_node_id = successors[0]
        else:
            # multiple successors exist
            # find the successor which represents the path decision included in the model
            suitable_successors = [succ for succ in successors if succ in model.path_decisions]
            if len(suitable_successors) != 1:
                requirements: List[int] = []
                for dec in model.path_decisions:
                    requirements += [r for r in get_requirements(graph, dec) if r not in requirements]
                if len(requirements) == 0:
                    # select the sequential version
                    for succ in successors:
                        if data_at(graph, succ).represents_sequential_version() and data_at(graph, succ).device_id in [
                            None,
                            experiment.get_system().get_host_device_id(),
                        ]:
                            suitable_successors = [succ]
                            model.path_decisions.append(succ)
                            break
                else:
                    # select the correct successor
                    for succ in successors:
                        if succ in requirements:
                            suitable_successors = [succ]
                            model.path_decisions.append(succ)
                            break
                if len(suitable_successors) != 1:
                    print("No correction possible")
                    show(graph)

                    raise ValueError(
                        "Invalid amount of potential successors (",
                        len(suitable_successors),
                        ") for path split at node:",
                        node_id,
                        "using decisions: ",
                        model.path_decisions,
                        "successors:",
                        successors,
                    )
            # suitable successor identified.
            # pass the current context to the successor
            next_node_id = suitable_successors[0]
    return context


def get_path_context(
    node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject, experiment
) -> ContextObject:
    """passes the context Object along the path and returns the context once the end has been reached"""
    # push device id to stack if necessary
    node_data = data_at(graph, node_id)
    if node_data.device_id is not None:
        context.last_seen_device_ids.append(node_data.device_id)

    # calculate context modifications for the current node
    context = __check_current_node(node_id, graph, model, context, experiment)

    # calculate context modifications for the children of the current node
    context = __check_children(node_id, graph, model, context, experiment)

    # pop device id from stack if necessary
    if node_data.device_id is not None:
        context.last_seen_device_ids.pop()

    # set last_visited_node_id to the original node_id,
    # since the calculation continues from node_id after the children have been visited
    context.last_visited_node_id = node_id

    # pass context to the right successor of the current node
    # At most a single node can be a successor, since the given model represents a single path through the graph.
    successors = get_successors(graph, node_id)
    if len(successors) == 1:
        # pass context to the single successor
        return get_path_context(successors[0], graph, model, context, experiment)

    elif len(successors) == 0:
        # no successor exists, return the current context
        return context

    else:
        # multiple successors exist
        # find the successor which represents the path decision included in the model
        suitable_successors = [succ for succ in successors if succ in model.path_decisions]
        if len(suitable_successors) != 1:
            requirements: List[int] = []
            for dec in model.path_decisions:
                requirements += [r for r in get_requirements(graph, dec) if r not in requirements]
            if len(requirements) == 0:
                # select the sequential version
                for succ in successors:
                    if data_at(graph, succ).represents_sequential_version() and data_at(graph, succ).device_id in [
                        None,
                        experiment.get_system().get_host_device_id(),
                    ]:
                        suitable_successors = [succ]
                        model.path_decisions.append(succ)
                        break
            else:
                # select the correct successor
                for succ in successors:
                    if succ in requirements:
                        suitable_successors = [succ]
                        model.path_decisions.append(succ)
                        break
            if len(suitable_successors) != 1:
                print("No correction possible")
                show(graph)

                raise ValueError(
                    "Invalid amount of potential successors (",
                    len(suitable_successors),
                    ") for path split at node:",
                    node_id,
                    "using decisions: ",
                    model.path_decisions,
                    "successors:",
                    successors,
                )
        # suitable successor identified.
        # pass the current context to the successor
        return get_path_context(suitable_successors[0], graph, model, context, experiment)


def __check_current_node(
    node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject, experiment
) -> ContextObject:
    """Check if the given node results in modifications to the given context.
    Return a modified version of the context which contains the required updates."""
    # due to the Read-Compute-Write paradigm used to create the Computational Units,
    # this structure is assumed for the nodes and their MemoryAccesses as well.

    # check if any data needs to be updated from a different device before reading
    # if so, the context will be supplemented with the identified updates
    node_data = data_at(graph, node_id)

    if isinstance(node_data, ContextNode):
        # apply modifications according to encountered context node
        updated_context = cast(ContextNode, data_at(graph, node_id)).get_modified_context(
            node_id, graph, model, context
        )
        return updated_context

    # only allow updates on device switches
    device_switch = False
    if data_at(graph, node_id).device_id is None:
        if context.last_seen_device_ids[-1] != context.last_visited_device_id:
            device_switch = True
    elif context.last_visited_device_id != data_at(graph, node_id).device_id:
        device_switch = True

    if device_switch:
        context = context.calculate_and_perform_necessary_updates(
            node_data.read_memory_regions,
            cast(int, context.last_seen_device_ids[-1]),
            node_data.node_id,
            graph,
            experiment,
        )

    # add the writes performed by the given node to the context
    context = context.add_writes(node_data.written_memory_regions, cast(int, context.last_seen_device_ids[-1]))

    return context


def __check_children(
    node_id: int, graph: nx.DiGraph, model: CostModel, context: ContextObject, experiment
) -> ContextObject:
    # pass context to all children
    for child in get_children(graph, node_id):
        # reset last_visited_node_id inbetween visiting children
        context.last_visited_node_id = node_id
        context = get_path_context_iterative(child, graph, model, context, experiment)
    return context



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import matplotlib.pyplot as plt  # type: ignore
import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.classes.edges.ChildEdge import ChildEdge
from discopop_library.discopop_optimizer.classes.edges.OptionEdge import OptionEdge
from discopop_library.discopop_optimizer.classes.edges.RequirementEdge import RequirementEdge
from discopop_library.discopop_optimizer.classes.edges.SuccessorEdge import SuccessorEdge
from discopop_library.discopop_optimizer.classes.edges.TemporaryEdge import TemporaryEdge
from discopop_library.discopop_optimizer.classes.nodes.ContextNode import ContextNode
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at

from networkx.drawing.nx_pydot import graphviz_layout  # type: ignore


def show(graph):
    """Plots the graph

    :return:
    """
    fig, ax = plt.subplots()
    try:
        # pos = nx.planar_layout(graph)  # good
        pos = graphviz_layout(graph, prog="dot")
    except nx.exception.NetworkXException:
        try:
            # fallback layouts
            pos = nx.shell_layout(graph)  # maybe
        except nx.exception.NetworkXException:
            pos = nx.random_layout(graph)

    drawn_nodes = set()
    nodes_lists = dict()
    node_ids = dict()
    node_insertion_sequence = []
    # draw nodes
    node_insertion_sequence.append(FunctionRoot)
    nodes_lists[FunctionRoot] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="#ff5151",
        node_shape="d",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), FunctionRoot)],
    )
    node_ids[FunctionRoot] = [n for n in graph.nodes if isinstance(data_at(graph, n), FunctionRoot)]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), FunctionRoot)])

    node_insertion_sequence.append(Loop)
    nodes_lists[Loop] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="#ff5151",
        node_shape="s",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), Loop)],
    )
    node_ids[Loop] = [n for n in graph.nodes if isinstance(data_at(graph, n), Loop)]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), Loop)])

    node_insertion_sequence.append(ContextNode)
    nodes_lists[ContextNode] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="yellow",
        node_shape="s",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), ContextNode)],
    )
    node_ids[ContextNode] = [n for n in graph.nodes if isinstance(data_at(graph, n), ContextNode)]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), ContextNode)])

    node_insertion_sequence.append(Workload)
    nodes_lists[Workload] = nx.draw_networkx_nodes(
        graph,
        pos=pos,
        ax=ax,
        node_size=200,
        node_color="#2B85FD",
        node_shape="o",
        nodelist=[n for n in graph.nodes if isinstance(data_at(graph, n), Workload) and n not in drawn_nodes],
    )
    node_ids[Workload] = [n for n in graph.nodes if isinstance(data_at(graph, n), Workload) and n not in drawn_nodes]
    drawn_nodes.update([n for n in graph.nodes if isinstance(data_at(graph, n), Workload) and n not in drawn_nodes])

    # id as label
    labels = {}
    for n in graph.nodes:
        labels[n] = graph.nodes[n]["data"].get_plot_label()
    nx.draw_networkx_labels(graph, pos, labels, font_size=7, ax=ax)

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="black",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], SuccessorEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="red",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], ChildEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="green",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], TemporaryEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="pink",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], OptionEdge)],
    )

    nx.draw_networkx_edges(
        graph,
        pos,
        ax=ax,
        edge_color="yellow",
        edgelist=[e for e in graph.edges(data="data") if isinstance(e[2], RequirementEdge)],
    )

    # define tool tip style when hovering
    # based on https://stackoverflow.com/questions/61604636/adding-tooltip-for-nodes-in-python-networkx-graph
    annot = ax.annotate(
        "",
        xy=(0, 0),
        xytext=(20, 20),
        textcoords="offset points",
        bbox=dict(boxstyle="round", fc="w"),
        arrowprops=dict(arrowstyle="->"),
    )
    annot.set_visible(False)

    idx_to_node_dict = {}
    for idx, node in enumerate(graph.nodes):
        idx_to_node_dict[idx] = node

    def update_annot(ind, node_ids_list):
        node_idx = ind["ind"][0]
        node_id = node_ids_list[node_idx]
        xy = pos[node_id]
        annot.xy = xy
        node_attr = {"node": node_id}
        node_attr.update(graph.nodes[node_id])
        text = data_at(graph, node_id).get_hover_text()
        annot.set_text(text)

    def hover(event):
        vis = annot.get_visible()
        if event.inaxes == ax:
            for node_type in node_insertion_sequence:
                nodes = nodes_lists[node_type]
                try:
                    cont, ind = nodes.contains(event)
                    if cont:
                        update_annot(ind, node_ids[node_type])
                        annot.set_visible(True)
                        fig.canvas.draw_idle()
                    else:
                        if vis:
                            annot.set_visible(False)
                            fig.canvas.draw_idle()
                except TypeError:
                    pass

    fig.canvas.mpl_connect("motion_notify_event", hover)
    plt.show()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


import os.path
from typing import Dict, Set, cast

from discopop_explorer.PEGraphX import MemoryRegion


def get_sizes_of_memory_regions(
    mem_regs: Set[MemoryRegion], mem_reg_file: str, return_all_memory_regions: bool = False
) -> Dict[MemoryRegion, int]:
    """Returns the size of the allocated memory region in bytes"""
    if not os.path.exists(mem_reg_file):
        raise ValueError("Unknown file: ", mem_reg_file)
    # load memory regions to dictionary
    mem_regs_dict: Dict[str, int] = dict()
    with open(mem_reg_file, "r") as f:
        for line in f.readlines():
            line = line.replace("\n", "")
            split_line = line.split(" ")
            mem_regs_dict[split_line[0]] = int(split_line[2])

    if return_all_memory_regions:
        # skip filtering of memory regions
        return cast(Dict[MemoryRegion, int], mem_regs_dict)

    result_dict: Dict[MemoryRegion, int] = dict()
    for mem_reg in mem_regs:
        if str(mem_reg) in mem_regs_dict:
            result_dict[mem_reg] = mem_regs_dict[str(mem_reg)]

    return result_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import jsons  # type: ignore

from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_library.CodeGenerator.classes.ContentBuffer import ContentBuffer
from discopop_library.CodeGenerator.classes.UnpackedSuggestion import UnpackedSuggestion
from discopop_library.discopop_optimizer.classes.system.devices.DeviceTypeEnum import DeviceTypeEnum
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID


def from_pattern_info(
    file_mapping: Dict[int, Path],
    patterns_by_type: Dict[str, List[PatternInfo]],
    skip_compilation_check: bool = False,
    compile_check_command: Optional[str] = None,
    CC="clang",
    CXX="clang++",
) -> Dict[int, str]:
    """Insert the given parallel patterns into the original source code.
    Returns a dictionary which maps the ID of every modified file to the updated contents of the file.
    This method does not modify the original source code.
    Only fileIDs of files which would be modified occur as keys in the returned dictionary.
    For now, it is assumed that the given patterns are independent of each other, i.e. no nesting exists.
    """
    # convert patterns to json strings so that only a single interface has to be maintained
    pattern_json_strings_by_type: Dict[str, List[str]] = dict()
    for type_str in patterns_by_type:
        if type_str not in pattern_json_strings_by_type:
            pattern_json_strings_by_type[type_str] = []
        for pattern in patterns_by_type[type_str]:
            pattern_json_strings_by_type[type_str].append(pattern.to_json())
    return from_json_strings(
        file_mapping,
        pattern_json_strings_by_type,
        skip_compilation_check=skip_compilation_check,
        compile_check_command=compile_check_command,
    )


def from_json_strings(
    file_mapping: Dict[int, Path],
    pattern_json_strings_by_type: Dict[str, List[str]],
    skip_compilation_check: bool = False,
    compile_check_command: Optional[str] = None,
    CC="clang",
    CXX="clang++",
) -> Dict[int, str]:
    """Insert the parallel patterns described by the given json strings into the original source code.
    Returns a dictionary which maps the ID of every modified file to the updated contents of the file.
    This method does not modify the original source code.
    Only fileIDs of files which would be modified occur as keys in the returned dictionary.
    For now, it is assumed that the given patterns are independent of each other, i.e. no nesting exists.
    """
    # convert pattern_json_strings to UnpackedSuggestion object
    unpacked_suggestions: List[UnpackedSuggestion] = []
    pattern_precedence = [
        "do_all",
        "reduction",
        "device_update",
        "gpu_do_all",
        "gpu_reduction",
        "pipeline",
        "simple_gpu",
        "combined_gpu",
    ]
    for type_str in pattern_precedence:
        if type_str not in pattern_json_strings_by_type:
            continue
        for json_str in pattern_json_strings_by_type[type_str]:
            unpacked_suggestions.append(UnpackedSuggestion(type_str, jsons.loads(json_str)))

    # create a dictionary mapping fileIds to ContentBuffer elements
    file_id_to_content_buffer: Dict[int, ContentBuffer] = dict()
    # iterate over UnpackedSuggestion objects
    for suggestion in unpacked_suggestions:
        # create new ContentBuffer if required
        if suggestion.file_id not in file_id_to_content_buffer:
            file_id_to_content_buffer[suggestion.file_id] = ContentBuffer(
                suggestion.file_id, file_mapping[suggestion.file_id]
            )
        # get pragma objects represented by the UnpackedSuggestion
        pragmas = suggestion.get_pragmas()
        # add pragmas to the ContentBuffer object which corresponds to the fileId specified in UnpackedSuggestions
        for pragma in pragmas:
            successful = file_id_to_content_buffer[suggestion.file_id].add_pragma(
                file_mapping,
                pragma,
                [],
                skip_compilation_check=skip_compilation_check,
                compile_check_command=compile_check_command,
                CC=CC,
                CXX=CXX,
            )
            # if the addition resulted in a non-compilable file, add the pragma as a comment
            if not successful:
                file_id_to_content_buffer[suggestion.file_id].add_pragma(
                    file_mapping,
                    pragma,
                    [],
                    add_as_comment=True,
                    compile_check_command=compile_check_command,
                    CC=CC,
                    CXX=CXX,
                )

    # create the resulting dictionary by assembling the contents of the ContentBuffer objects
    result_dict: Dict[int, str] = dict()
    for file_id in file_id_to_content_buffer:
        result_dict[file_id] = file_id_to_content_buffer[file_id].get_modified_source_code()
    return result_dict


def from_json_strings_with_mapping(
    file_mapping: Dict[int, Path],
    pattern_json_strings_with_mapping_by_type: Dict[str, List[Tuple[str, DeviceID, Optional[DeviceTypeEnum]]]],
    skip_compilation_check: bool = False,
    compile_check_command: Optional[str] = None,
    CC="clang",
    CXX="clang++",
    host_device_id=0,
) -> Dict[int, str]:
    """Insert the parallel patterns described by the given json strings into the original source code.
    Returns a dictionary which maps the ID of every modified file to the updated contents of the file.
    This method does not modify the original source code.
    Only fileIDs of files which would be modified occur as keys in the returned dictionary.
    For now, it is assumed that the given patterns are independent of each other, i.e. no nesting exists.
    """
    # convert pattern_json_strings to UnpackedSuggestion object
    unpacked_suggestions: List[UnpackedSuggestion] = []
    pattern_precedence = [
        "device_update",
        "do_all",
        "reduction",
        "gpu_do_all",
        "gpu_reduction",
        "pipeline",
        "simple_gpu",
        "combined_gpu",
    ]
    for type_str in pattern_precedence:
        if type_str not in pattern_json_strings_with_mapping_by_type:
            continue
        for json_str, device_id, device_type in pattern_json_strings_with_mapping_by_type[type_str]:
            unpacked_suggestions.append(
                UnpackedSuggestion(
                    type_str,
                    jsons.loads(json_str),
                    device_id=device_id,
                    device_type=device_type,
                    host_device_id=host_device_id,
                )
            )

    # create a dictionary mapping fileIds to ContentBuffer elements
    file_id_to_content_buffer: Dict[int, ContentBuffer] = dict()
    # iterate over UnpackedSuggestion objects
    for suggestion in unpacked_suggestions:
        # create new ContentBuffer if required
        if suggestion.file_id not in file_id_to_content_buffer:
            file_id_to_content_buffer[suggestion.file_id] = ContentBuffer(
                suggestion.file_id, file_mapping[suggestion.file_id]
            )
        # get pragma objects represented by the UnpackedSuggestion
        pragmas = suggestion.get_pragmas()
        # add pragmas to the ContentBuffer object which corresponds to the fileId specified in UnpackedSuggestions
        for pragma in pragmas:
            successful = file_id_to_content_buffer[suggestion.file_id].add_pragma(
                file_mapping,
                pragma,
                [],
                skip_compilation_check=skip_compilation_check,
                compile_check_command=compile_check_command,
                CC=CC,
                CXX=CXX,
            )
            # if the addition resulted in a non-compilable file, add the pragma as a comment
            if not successful:
                file_id_to_content_buffer[suggestion.file_id].add_pragma(
                    file_mapping,
                    pragma,
                    [],
                    add_as_comment=True,
                    compile_check_command=compile_check_command,
                    CC=CC,
                    CXX=CXX,
                )

    # create the resulting dictionary by assembling the contents of the ContentBuffer objects
    result_dict: Dict[int, str] = dict()
    for file_id in file_id_to_content_buffer:
        result_dict[file_id] = file_id_to_content_buffer[file_id].get_modified_source_code()
    return result_dict



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os
from enum import IntEnum
from typing import List, Set, Optional, Union, Any, Dict, Tuple, cast

from discopop_explorer.PEGraphX import (
    PEGraphX,
    CUNode,
    parse_id,
    DepType,
    NodeID,
    LineID,
    MemoryRegion,
    FunctionNode,
    Node,
    LoopNode,
)
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.utils import (
    __get_dep_of_type as get_dep_of_type,
)
from discopop_explorer.utils import (
    is_scalar_val,
    is_loop_index2,
    classify_loop_variables,
    get_initialized_memory_regions_in,
)
from discopop_explorer.utils import (
    is_written_in_subtree,
    is_func_arg,
    is_readonly,
    is_global,
    is_read_in_right_subtree,
    is_first_written,
    is_read_in_subtree,
)
from discopop_explorer.variable import Variable
from discopop_library.MemoryRegions.utils import get_sizes_of_memory_regions
from .GPUMemory import getCalledFunctions, map_node

"""
def remove_duplicates(my_list: List) -> List:
    ""removes the duplicates of a given list

    :param my_list:
    :return:
    ""
    return list(dict.fromkeys(my_list))
"""


def fmt_vars(v: List[Variable]) -> str:
    """inserts the elements of the VarID vector 'v' into 'ss';
        elements are separated by ','

    :param v:
    :return:
    """
    if not v:  # if (v.empty())
        return ""

    ss: str = v[0].name
    for i in range(1, len(v)):
        ss += ", " + v[i].name
    return ss


def omp_clause_str(name: str, args: List[str]) -> str:
    """generates a string for omp clauses/constructs

    :param name: contains 'args' if args list is not empty
    :param args:
    :return:
    """

    # check if "args" contained in name
    if "args" in name and len(args) > 0:
        # replace "args" with contents of args list
        return name.replace("args", ",".join(args))
    elif len(args) == 0:
        return name
    else:
        raise ValueError("Unsupported name: ", name, "and args:", args)


def omp_clause_str_old(name: str, args: List[str]) -> str:
    """generates a json-string for omp clauses/constructs

    :param name:
    :param args:
    :return:
    """
    # name, converted std::ostringstream to str
    result: str = '{"name":"' + name + '",'

    # args
    result += '"args":['
    if args:
        result += '"' + args[0] + '"'
        for i in range(1, len(args)):
            result += ',"' + args[i] + '"'
    result += "]}"
    return result


def omp_construct_str(name: str, line: int, clauses: List[str]) -> str:
    """

    :param name:
    :param line:
    :param clauses:
    :return:
    """
    result: str = '{"name":"' + name + '",'
    result += '"line":' + str(line) + ","
    result += '"clauses":['
    if clauses:
        result += clauses[0]
    for i in range(1, len(clauses)):
        result += "," + clauses[i]
    result += "]}"
    return result


class OmpConstructPositioning(IntEnum):
    BEFORE_LINE = 0
    AFTER_LINE = 1


def omp_construct_dict(
    name: str,
    line: LineID,
    clauses: List[str],
    positioning: OmpConstructPositioning = OmpConstructPositioning.BEFORE_LINE,
) -> Dict[str, Union[str, LineID, List[str], OmpConstructPositioning]]:
    """

    :param name:
    :param line:
    :param clauses:
    :return:
    """
    result: Dict[str, Union[str, LineID, List[str], OmpConstructPositioning]] = dict()
    result["name"] = name
    result["line"] = line
    result["clauses"] = clauses
    result["positioning"] = positioning
    return result


# inherits from class ParallelPattern (=PatternInfo)
class GPULoopPattern(PatternInfo):
    # public:
    called_functions: Set[NodeID]
    map_type_to: List[Variable]
    map_type_from: List[Variable]
    map_type_tofrom: List[Variable]
    map_type_alloc: List[Variable]
    reduction_vars_str: List[str]
    reduction_vars_ids: List[Variable]
    iteration_count: int = 0
    has_scalar_reduction_var: bool
    nodeID: NodeID
    # new
    nestedLoops: Set[NodeID]
    nextLoop: Optional[NodeID]
    parentLoop: str
    collapse: int
    scheduling: str
    constructs: List[Dict[Any, Any]]
    project_folder_path: str

    def __init__(
        self,
        pet: PEGraphX,
        nodeID: NodeID,
        startLine,
        endLine,
        iterationCount: int,
        project_folder_path: str,
        reduction_vars: Optional[List[Variable]] = None,
    ):
        node = map_node(pet, nodeID)
        super().__init__(node)  # PatternInfo(node)
        PatternInfo.iterations_count = iterationCount
        self.has_scalar_reduction_var = False  # copy.deepcopy(False)
        self.nodeID = nodeID
        self.startLine = startLine
        self.endLine = endLine
        self.iterationCount = iterationCount
        self.project_folder_path = project_folder_path
        # explicitly initialize empty, else it will copy values of other patterns
        self.map_type_to = []
        self.map_type_tofrom = []
        self.map_type_alloc = []
        self.map_type_from = []
        # should also be done with all others? probably
        self.nestedLoops = set()
        self.called_functions = set()
        if reduction_vars is None:
            self.reduction_vars_str: List[str] = []
            self.reduction_vars_ids: List[Variable] = []
        else:
            self.reduction_vars_str: List[str] = [cast(str, v.operation) + ":" + v.name for v in reduction_vars]
            self.reduction_vars_ids: List[Variable] = reduction_vars
        self.iteration_count = 0
        self.nextLoop = None
        self.parentLoop = ""
        self.collapse = 1
        self.scheduling = ""
        self.constructs = []
        self.declared_global_variables: Set[Variable] = set()

    def __str__(self):
        raise NotImplementedError()  # used to identify necessity to call to_string() instead

    def to_string(self, pet: PEGraphX) -> str:
        constructs = self.__get_constructs(pet, self.project_folder_path)
        construct_str = "\n" if len(constructs) > 0 else ""
        for entry in constructs:
            for key in entry:
                construct_str += "\t" + key + ": " + str(entry[key]) + "\n"
            construct_str += "\n"
        return (
            f"{'DoAll (GPU)' if not self.reduction_vars_str else 'Reduction (GPU)'} at: {self.node_id}\n"
            f"Start line: {self.startLine}\n"
            f"End line: {self.endLine}\n"
            f"Collapse: {self.collapse}\n"
            f"Map to: {self.map_type_to}\n"
            f"Map from: {self.map_type_from}\n"
            f"Map to_from: {self.map_type_tofrom}\n"
            f"Map alloc: {self.map_type_alloc}\n"
            f"Reduction vars: {self.reduction_vars_str}\n"
            f"OpenMP constructs: {construct_str}\n"
        )

    def save_omp_constructs(self, pet: PEGraphX, project_folder_path: str):
        """Save OpenMP constructs such that they are included in the exported JSON file."""
        constructs = self.__get_constructs(pet, project_folder_path)
        self.constructs = constructs

    def toJson(self, pet: PEGraphX, project_folder_path: str) -> str:
        """Generates a json-string which contains the information about how to
            implement this pattern using OpenMP constructs

        :return:
        """
        json_output: str = "{"
        # == == Metadata == ==
        json_output += '"id":"' + str(self.nodeID) + '",'
        json_output += '"startline":"' + self.start_line + '",'
        json_output += '"endline":"' + self.end_line + '",'

        # == == Type == ==
        json_type: str = "DoAll (GPU)" if not self.reduction_vars_str else "Reduction (GPU)"
        json_output += '"type":"' + json_type + '",'

        # == == Constructs == ==
        json_output += '"constructs":['

        constructs: List[Any] = self.__get_constructs(pet, project_folder_path)

        # add all constructs to the output string
        json_output += constructs[0]
        for i in range(1, len(constructs)):
            json_output += "," + constructs[i]

        json_output += "]}"
        return json_output

    def __get_constructs(
        self, pet: PEGraphX, project_folder_path: str
    ) -> List[Dict[str, Union[str, LineID, List[str], OmpConstructPositioning]]]:
        constructs: List[Dict[str, Union[str, LineID, List[str], OmpConstructPositioning]]] = []

        # == default construct ==
        clauses: List[str] = []
        var_names: List[str] = []
        modified_var_names: List[str] = []
        subnodes = pet.subtree_of_type(pet.node_at(self.node_id), CUNode)
        if self.map_type_to:
            modified_var_names = []
            for var in self.map_type_to:
                memory_regions = pet.get_memory_regions(subnodes, var.name)

                # get size of memory region
                memory_region_sizes = get_sizes_of_memory_regions(
                    memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
                )
                if len(memory_region_sizes) > 0:
                    max_mem_reg_size = max(memory_region_sizes.values())
                    # divide memory region size by size of variable
                    # construct new list of modified var names
                    modified_var_names.append(
                        (
                            # var.name + "[:]"
                            var.name + "[:" + str(int(max_mem_reg_size / var.sizeInByte)) + "]"
                            if "**" in var.type
                            else var.name
                        )
                    )
                else:
                    modified_var_names.append(var.name + "[:]" if "**" in var.type else var.name)

            clauses.append(omp_clause_str("map(to: args)", modified_var_names))
            var_names = []

        if self.map_type_from:
            modified_var_names = []
            for var in self.map_type_from:
                memory_regions = pet.get_memory_regions(subnodes, var.name)

                # get size of memory region
                memory_region_sizes = get_sizes_of_memory_regions(
                    memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
                )
                if len(memory_region_sizes) > 0:
                    max_mem_reg_size = max(memory_region_sizes.values())
                    # divide memory region size by size of variable
                    # construct new list of modified var names
                    modified_var_names.append(
                        (
                            var.name + "[:" + str(int(max_mem_reg_size / var.sizeInByte)) + "]"
                            if "**" in var.type
                            else var.name
                        )
                    )
                else:
                    modified_var_names.append(var.name + "[:]" if "**" in var.type else var.name)
            clauses.append(omp_clause_str("map(from: args)", modified_var_names))
            var_names = []

        if self.map_type_tofrom:
            modified_var_names = []
            for var in self.map_type_tofrom:
                memory_regions = pet.get_memory_regions(subnodes, var.name)

                # get size of memory region
                memory_region_sizes = get_sizes_of_memory_regions(
                    memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
                )
                if len(memory_region_sizes) > 0:
                    max_mem_reg_size = max(memory_region_sizes.values())
                    # divide memory region size by size of variable
                    # construct new list of modified var names
                    modified_var_names.append(
                        (
                            var.name + "[:" + str(int(max_mem_reg_size / var.sizeInByte)) + "]"
                            if "**" in var.type
                            else var.name
                        )
                    )
                else:
                    modified_var_names.append(var.name + "[:]" if "**" in var.type else var.name)
            clauses.append(omp_clause_str("map(tofrom: args)", modified_var_names))
            var_names = []

        if self.map_type_alloc:
            modified_var_names = []
            for var in self.map_type_alloc:
                memory_regions = pet.get_memory_regions(subnodes, var.name)

                # get size of memory region
                memory_region_sizes = get_sizes_of_memory_regions(
                    memory_regions, os.path.join(project_folder_path, "profiler/memory_regions.txt")
                )
                if len(memory_region_sizes) > 0:
                    max_mem_reg_size = max(memory_region_sizes.values())
                    # divide memory region size by size of variable
                    # construct new list of modified var names
                    modified_var_names.append(
                        (
                            var.name + "[:" + str(int(max_mem_reg_size / var.sizeInByte)) + "]"
                            if "**" in var.type
                            else var.name
                        )
                    )
                else:
                    modified_var_names.append(var.name + "[:]" if "**" in var.type else var.name)
            clauses.append(omp_clause_str("map(alloc: args)", modified_var_names))
            var_names = []

        if self.reduction_vars_str:
            clauses.append(omp_clause_str("reduction(args)", self.reduction_vars_str))
            if self.has_scalar_reduction_var:
                clauses.append(omp_clause_str("defaultmap(tofrom:scalar)", []))

        tmp_start_line = LineID(str(self._node.file_id) + ":" + str(self.startLine))
        constructs.append(
            omp_construct_dict("#pragma omp target teams distribute parallel for", tmp_start_line, clauses)
        )

        # == additional constructs ==
        used_global_vars: Set[Variable] = set()
        for node_id in self.called_functions:
            fn_node: FunctionNode = cast(FunctionNode, map_node(pet, node_id))
            fn_node_start_line = LineID(str(fn_node.file_id) + ":" + str(fn_node.start_line))
            fn_node_end_line = LineID(str(fn_node.file_id) + ":" + str(fn_node.end_line + 1))
            constructs.append(omp_construct_dict("#pragma omp declare target", fn_node_start_line, []))
            constructs.append(
                omp_construct_dict(
                    "#pragma omp end declare target",
                    fn_node_end_line,
                    [],
                    positioning=OmpConstructPositioning.AFTER_LINE,
                )
            )
            # get used global variables
            cu_nodes: List[CUNode] = pet.subtree_of_type(fn_node, CUNode)
            tmp_global_vars: Set[Variable] = set()
            for cu_node in cu_nodes:
                tmp_global_vars.update(cu_node.global_vars)

            # check if global var is defined outside fn_node's body and update used_global_vars
            used_global_vars.update(
                global_var
                for global_var in tmp_global_vars
                if global_var.defLine
                if not fn_node.contains_line(global_var.defLine)
            )

        # declare all global variables used in called functions
        self.declared_global_variables.update(used_global_vars)
        for global_var in used_global_vars:
            constructs.append(
                omp_construct_dict("#pragma omp declare target  // " + str(global_var.name), global_var.defLine, [])
            )
            constructs.append(
                omp_construct_dict(
                    "#pragma omp end declare target  // " + str(global_var.name),
                    global_var.defLine,
                    [],
                    positioning=OmpConstructPositioning.AFTER_LINE,
                )
            )

        return constructs

    def getDataStr(self, pet: PEGraphX) -> str:
        """Generates a string which contains data that is used to rank this loop, e.g.
            the number of iterations and which variables have to be transferred etc.

        :return:
        """
        # The string begins with the type, the LID and the number of iterations.
        fileID = parse_id(self.nodeID)
        ss: str = str(fileID[0]) + ":"
        ss += str(self.startLine) + " "
        ss += "0" if not self.reduction_vars_str else "1"
        ss += " " + str(self.iteration_count) + " "

        # This is followed by three arrays, each containing variable names separated
        # by spaces, e.g. [var_1 var_2 ... var_n].
        # The first array references the variables that have to be copied to the GPU,
        # the second array references the variables that have to be copied to host
        # memory, and the variable names in the third array correspond to the
        # variables that have to be copied to the GPU and then back to host memory.
        ss += "["
        ss += fmt_vars(self.map_type_to)
        ss += "] ["
        ss += fmt_vars(self.map_type_from)
        ss += "] ["
        tmp: List[Variable] = self.map_type_tofrom

        for va in self.reduction_vars_ids:
            tmp.append(va)
        ss += fmt_vars(tmp)
        ss += "]"

        # The final part of this string contains information about the loop's nested
        # loops including their iteration numbers.
        n: LoopNode = cast(LoopNode, map_node(pet, self.nodeID))
        total_i: int = n.loop_iterations
        for cn_id in pet.direct_children(n):
            if cn_id.type == 2:  # type loop
                ss += self.__add_sub_loops_rec(pet, cn_id.id, total_i)
        return ss

    def classifyLoopVars(self, pet: PEGraphX, loop: LoopNode) -> None:
        """Classify the variables that are accessed in this loop, e.g. assign them
            to a map-type vector and find reduction variables

        :return:
        """
        reduction = []
        lst = pet.get_left_right_subtree(loop, False)
        rst = pet.get_left_right_subtree(loop, True)
        sub = pet.subtree_of_type(loop, CUNode)

        raw = set()
        war = set()
        waw = set()
        rev_raw = set()

        dummyFunctions: Set[NodeID] = set()
        self.called_functions.update(getCalledFunctions(pet, loop, self.called_functions, dummyFunctions))

        for sub_node in sub:
            raw.update(get_dep_of_type(pet, sub_node, DepType.RAW, False))
            war.update(get_dep_of_type(pet, sub_node, DepType.WAR, False))
            waw.update(get_dep_of_type(pet, sub_node, DepType.WAW, False))
            rev_raw.update(get_dep_of_type(pet, sub_node, DepType.RAW, True))

        # global vars need to be considered as well since mapping / updates may be required
        vars = pet.get_undefined_variables_inside_loop(loop, include_global_vars=True)

        _, private_vars, _, _, _ = classify_loop_variables(pet, loop)

        # define temporary classification lists
        map_type_to: List[Tuple[Variable, Set[MemoryRegion]]] = []
        map_type_tofrom: List[Tuple[Variable, Set[MemoryRegion]]] = []
        map_type_from: List[Tuple[Variable, Set[MemoryRegion]]] = []
        map_type_alloc: List[Tuple[Variable, Set[MemoryRegion]]] = []

        for var in vars:
            if var in private_vars:
                continue
            if is_scalar_val(var) and var.accessMode == "R" and False:
                # will be implicitly mapped as firstprivate
                continue
            if is_loop_index2(pet, loop, var.name):
                continue
            elif loop.reduction and pet.is_reduction_var(loop.start_position(), var.name):
                var.operation = pet.get_reduction_sign(loop.start_position(), var.name)
                reduction.append(var)
            # TODO grouping

            if (
                is_written_in_subtree(vars[var], raw, waw, lst)
                or is_func_arg(pet, var.name, loop)
                or (
                    # "manual" triggering of "map(to)" required for true global variables since initialization of global
                    # variables might not occur in dependencies since the initializations are not instrumented
                    is_global(var.name, sub)
                    and not pet.get_parent_function(loop).contains_line(var.defLine)
                )
            ):
                if is_readonly(vars[var], war, waw, rev_raw):
                    map_type_to.append((var, vars[var]))
                elif is_read_in_right_subtree(vars[var], rev_raw, sub):
                    map_type_tofrom.append((var, vars[var]))
                elif is_written_in_subtree(vars[var], raw, waw, cast(List[Node], sub)):
                    map_type_alloc.append((var, vars[var]))
                else:
                    pass

            elif is_first_written(vars[var], raw, war, sub):
                # TODO simplify
                if is_read_in_subtree(vars[var], rev_raw, rst):
                    map_type_from.append((var, vars[var]))
                else:
                    map_type_alloc.append((var, vars[var]))
            else:
                pass

        # use known variables to reconstruct the correct variable names from the classified memory regions
        left_subtree_without_called_nodes = pet.get_left_right_subtree(loop, False, ignore_called_nodes=True)
        prior_known_vars = pet.get_variables(left_subtree_without_called_nodes)
        # get memory regions which are initialized in the loop and treat them like prior known vars wrt. de-aliasing
        initilized_in_loop = get_initialized_memory_regions_in(pet, sub)
        combined_know_vars: Dict[Variable, Set[MemoryRegion]] = dict()
        for var in prior_known_vars:
            if var not in combined_know_vars:
                combined_know_vars[var] = set()
            combined_know_vars[var].update(prior_known_vars[var])
        for var in initilized_in_loop:
            if var not in combined_know_vars:
                combined_know_vars[var] = set()
            combined_know_vars[var].update(initilized_in_loop[var])

        # de-alias and store identified mapping information
        self.map_type_to = self.__apply_dealiasing(map_type_to, combined_know_vars)
        self.map_type_tofrom = self.__apply_dealiasing(map_type_tofrom, combined_know_vars)

        self.map_type_alloc = self.__apply_dealiasing(map_type_alloc, combined_know_vars)

        self.map_type_from = self.__apply_dealiasing(map_type_from, combined_know_vars)

    def __apply_dealiasing(
        self,
        input_list: List[Tuple[Variable, Set[MemoryRegion]]],
        previously_known: Dict[Variable, Set[MemoryRegion]],
    ) -> List[Variable]:
        """Apply de-aliasing such that only valid (i.e. known) variable names are returned.
        Memory Regions specified in initialized_memory_regions will be ignored respectively passed through.
        """

        tmp_memory_regions = set()
        for _, mem_regs in input_list:
            tmp_memory_regions.update(mem_regs)
        cleaned = [pkv for pkv in previously_known if len(previously_known[pkv].intersection(tmp_memory_regions))]
        return cleaned

    def setParentLoop(self, pl: str) -> None:
        """

        :param pl:
        :return:
        """
        self.parentLoop = pl

    def getNestedLoops(self, pet: PEGraphX, node_id: NodeID) -> None:
        """

        :param node_id:
        :return:
        """
        # calculate the number of iterations of this loop relative to the top loop
        n = map_node(pet, node_id)

        # extend the string stream with this information and scan all child nodes to
        # identify and process further nested loops
        for cn_id in pet.direct_children(n):
            if cn_id.type == 2:  # type loop
                self.nestedLoops.add(cn_id.id)
                self.getNestedLoops(pet, cn_id.id)

    def getNextLoop(self, pet: PEGraphX, node_id: NodeID) -> None:
        """

        :param node_id:
        :return:
        """
        n = map_node(pet, node_id)
        endLine = 0
        for children in pet.direct_children(n):
            if children.end_line > endLine:
                endLine = children.end_line

    def setCollapseClause(self, pet: PEGraphX, node_id: NodeID, res):
        """

        :param node_id:
        :return:
        """
        # calculate the number of iterations of this loop relative to the top loop
        n: LoopNode = cast(LoopNode, map_node(pet, node_id))

        do_all_loops = [node.node_id for node in res.patterns.do_all]

        loop_entry_node = cast(LoopNode, pet.node_at(node_id)).get_entry_node(pet)
        if loop_entry_node is None:
            loop_entry_node = pet.direct_children(pet.node_at(node_id))[0]

        for cn_id in pet.direct_children(n):
            if cn_id.type == 2:  # check for loop node contained in the loop body
                if cn_id.end_line <= n.end_line:  # todo not true if loop bodies are terminated by braces
                    # only consider child as collapsible, if it is a do-all loop
                    if cn_id.id in do_all_loops:
                        # check for perfect nesting of both loops (i.e. no statements inbetween)
                        # todo: possible improvement: consider columns as well, or test using the AST / nesting
                        #  information from the LLVM debug information
                        # check if distance between first CU of node_id and cn_id is 2 steps on the successor graph
                        potentials: Set[Node] = set()
                        for succ1 in pet.direct_successors(loop_entry_node):
                            for succ2 in pet.direct_successors(succ1):
                                potentials.add(succ2)
                        if cast(LoopNode, cn_id).get_entry_node(pet) in potentials:
                            # perfect nesting possible. allow collapsing the loops, if root loop has no other children
                            if len(pet.direct_children(pet.node_at(node_id))) == 5:
                                # 2 children for loop condition and body
                                # 1 child is the collapsible loop
                                # 2 children for loop end and increment
                                self.collapse += 1
                                self.setCollapseClause(pet, cn_id.id, res)

    def findMappedVar(self, direction: str, var: Variable) -> bool:
        """

        :param var:
        :param direction:
        :return:
        """
        if direction == "alloc":
            for v in self.map_type_alloc:
                if v == var:
                    return True
        elif direction == "to":
            for v in self.map_type_to:
                if v == var:
                    return True
        elif direction == "from":
            for v in self.map_type_from:
                if v == var:
                    return True
        elif direction == "tofrom":
            for v in self.map_type_tofrom:
                if v == var:
                    return True
        return False

    def printGPULoop(self) -> None:
        """

        :return:
        """
        print("#pragma omp target data ")
        print(self.node_id + " = " + self.start_line + " " + self.end_line)
        # long ll = determineLineNumber(firstGPULoop.getStartLine());
        print("map_type_alloc: ")
        for k in [v.name for v in self.map_type_alloc]:
            print("    " + k)
        print("map_type_to: ")
        for k in [v.name for v in self.map_type_to]:
            print("    " + k)
        print("map_type_from: ")
        for k in [v.name for v in self.map_type_from]:
            print("    " + k)
        print("map_type_tofrom: ")
        for k in [v.name for v in self.map_type_tofrom]:
            print("    " + k)

    def __add_sub_loops_rec(self, pet: PEGraphX, node_id: NodeID, top_loop_iterations: int) -> str:
        """This function adds information about the loop's child loops to the string
            stream 'ss'. This information contains the child loop's line number and
            its number of iterations divided by the number of iterations of the top loop.
            For each loop, the string is extended with " line_number iteration_count".

        :param node_id:
        :param top_loop_iterations:
        :return:
        """
        # calculate the number of iterations of this loop relative to the top loop
        n = map_node(pet, node_id)
        ll = n.start_line
        total_i = cast(LoopNode, n).loop_iterations
        i_cnt = 0 if top_loop_iterations == 0 else total_i / top_loop_iterations

        # extend the string stream with this information and scan all child nodes to
        # identify and process further nested loops
        ss: str = " " + str(ll) + "-" + str(i_cnt)
        for cn_id in pet.direct_children(n):
            if cn_id.type == 2:
                ss += self.__add_sub_loops_rec(pet, cn_id.id, top_loop_iterations)
        return ss



from discopop_explorer.pattern_detectors.do_all_detector import DoAllInfo


class DoAllInfoForValidation(object):
    dai: DoAllInfo

    def __init__(self, do_all_info: DoAllInfo):
        self.dai = do_all_info

    def __eq__(self, other):
        if (
            self.dai.start_line == other.dai.start_line
            and self.dai.end_line == other.dai.end_line
            and self.dai.reduction == other.dai.reduction
            and self.dai.shared == other.dai.shared
            and self.dai.private == other.dai.private
            and self.dai.first_private == other.dai.first_private
            and self.dai.last_private == other.dai.last_private
        ):
            return True
        return False



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import json
import sys
from dataclasses import dataclass
from typing import List


@dataclass
class PatchApplicatorArguments(object):
    """Container Class for the arguments passed to the discopop_patch_applicator"""

    verbose: bool
    apply: List[str]
    rollback: List[str]
    clear: bool
    load: bool
    list: bool

    def __post_init__(self):
        self.__validate()

    def __validate(self):
        """Validate the arguments passed to the discopop_patch_applicator, e.g check if given files exist"""
        # check mutually exclusive arguments
        exit_required = False
        if self.clear and self.load:
            print("Please use only one of `--clear` or `--load`.")
            exit_required = True
        if len(self.apply) > 0 and len(self.rollback) > 0:
            print("Please use only one of '--apply' or '--rollback'.")
            exit_required = True
        if self.clear and (len(self.apply) > 0 or len(self.rollback)):
            print("Please use either '--clear' or any of '--apply' or '--rollback'.")
            exit_required = True
        if self.load and (len(self.apply) > 0 or len(self.rollback)):
            print("Please use either '--load' or any of '--apply' or '--rollback'.")
            exit_required = True
        if self.list:
            self.apply = []
            self.rollback = []
            self.clear = False
            self.load = False
        if exit_required:
            print("Exiting.")
            sys.exit(0)
